UNIT Linker;

{$mode objfpc}{$H+}

INTERFACE

USES
  Classes, Contnrs, SysUtils, Dialogs, Base;

CONST
  versionkey = #1;
  MT         = 12;
  MTOrg      = $20;
  DescSize   = 80;

TYPE
  Module = CLASS;
  ModuleName = ARRAY [0..32-1] OF CHAR;
  Buffer = ARRAY [0..63 * 1024 DIV 4 - 1] OF INTEGER;

  ImageModDesc = RECORD
    name: ModuleName;
    next: INTEGER;
    key, num, size, refcnt: INTEGER;
    data, code, imp, cmd, ent, ptr, unused: INTEGER;
  END;

  Module = CLASS
    next: Module;
    addr: INTEGER;
    size: INTEGER;
    desc: ImageModDesc;
  END;

VAR
  root: Module;
  AllocPtr, res, buffsize: INTEGER;
  importing, imported: ModuleName;
  GC: TObjectList;

  PROCEDURE LinkDiskImage(path,modname: ARRAY OF CHAR; corename: ARRAY OF CHAR);

IMPLEMENTATION

FUNCTION _DIV(CONST x, y: INTEGER): INTEGER;
VAR
  Q,R: INTEGER;
BEGIN
  Q:=TRUNC(x/y);
  R:=x MOD y;
  IF R<0 THEN BEGIN
    IF y>0 THEN
      Q:=Q-1
    ELSE
      Q:=Q+1;
  END;
  RESULT := Q;
END;

FUNCTION _MOD(CONST x, y: INTEGER): INTEGER;
BEGIN
  ASSERT(y > 0);
  IF x >= 0 THEN
    RESULT := x MOD y
  ELSE
    RESULT := (y + (x MOD y)) MOD y;
END;

PROCEDURE error(n: INTEGER; name: ARRAY OF CHAR);
BEGIN
  res := n;
  importing := name;
END;

PROCEDURE Check(s: ARRAY OF CHAR);
VAR
  i: INTEGER;
  ch: CHAR;
BEGIN
  ch := s[0];
  res := 1;
  i := 1;
  IF (ch >= 'A') AND (ch <= 'Z') OR (ch >= 'a') AND (ch <= 'z') THEN BEGIN
    REPEAT
      ch := s[i];
      INC(i);
    UNTIL NOT ((ch >= '0') AND (ch <= '9') OR (ch >= 'A') AND (ch <= 'Z')
      OR (ch >= 'a') AND (ch <= 'z') OR (ch = '.')) OR (i = 32);
    IF (i < 32) AND (ch = #0) THEN res := 0;
  END;
END;

PROCEDURE Read(VAR F: TFileStream; VAR ch: CHAR);
BEGIN
  F.Read(ch,1);
END;

PROCEDURE ReadInt(VAR F: TFileStream; VAR x: INTEGER);
BEGIN
  F.Read(x,4);
END;

PROCEDURE ReadString(VAR F: TFileStream; VAR x: ARRAY OF CHAR);
VAR
  i: INTEGER;
  ch: CHAR;
BEGIN
  i := 0;
  Read(F, ch);
  WHILE ch <> #0 DO BEGIN
    IF i < LENGTH(x)-1 THEN BEGIN
      x[i] := ch;
      INC(i)
    END;
    Read(F, ch);
  END;
  x[i] := #0;
END;

PROCEDURE ReadStringPart(VAR F: TFileStream; VAR n, cnt: INTEGER);
VAR
  ch: CHAR;
BEGIN
  n := 0;
  cnt := 0;
  Read(F, ch);
  WHILE (cnt < 4) AND (ch <> #0) DO BEGIN
    INC(n, (ORD(ch) SHL (cnt * 8)));
    INC(cnt);
    IF cnt < 4 THEN Read(F, ch);
  END;
END;

PROCEDURE Load(path, name: ARRAY OF CHAR; VAR buffer: ARRAY OF INTEGER; VAR newmod: Module);
VAR
  amod, impmod: Module;
  i, n, key, impkey, mno, nofimps, size: INTEGER;
  p, u, v, w: INTEGER;
  ch: CHAR;
  body: INTEGER;
  fixorgP, fixorgD, fixorgT: INTEGER;
  disp, adr, inst, pno, vno, dest, offset: INTEGER;
  name1, impname: ModuleName;
  F : TFileStream;
  import: ARRAY [0..16-1] OF Module;
BEGIN

  amod := root;
  res := 0;
  nofimps := 0;

  WHILE (amod <> NIL) AND (name <> amod.desc.name) DO amod := amod.next;

  IF amod = NIL THEN BEGIN(*load*)
    Check(name);
    F := TFileStream.Create(path+name+'.rsc',fmOpenRead);
    IF FileExists(path+name+'.rsc') THEN BEGIN
      F.Seek(0,soFromBeginning);
      ReadString(F, name1);
      ReadInt(F, key);
      Read(F, ch);
      ReadInt(F, size);
      importing := name1;
      IF ch = versionkey THEN BEGIN
        ReadString(F, impname);   (*imports*)
        WHILE (impname[0] <> #0) AND (res = 0) DO BEGIN
          ReadInt(F, impkey);
          Load(path,impname, buffer, impmod);
          import[nofimps] := impmod;
          importing := name1;
          IF res = 0 THEN BEGIN
            IF impmod.desc.key = impkey THEN BEGIN
              INC(impmod.desc.refcnt);
              INC(nofimps)
            END ELSE BEGIN
              error(3, name1);
              imported := impname;
            END;
          END;
          ReadString(F, impname)
        END;
      END ELSE
        error(2, name1);
    END ELSE
      error(1, name);

    IF res = 0 THEN BEGIN(*search for a hole in the list allocate and link*)
      INC(size, DescSize);
      IF AllocPtr + size < LENGTH(buffer) * 4 THEN BEGIN
        p := _DIV(AllocPtr,4);
        AllocPtr := _DIV(AllocPtr + size + $100,$20) * $20;
        amod := Module.Create;
        GC.Add(amod);
        amod.next := root;
        root := amod;
        amod.addr := p * 4;
        amod.desc.size := AllocPtr - p * 4;
        IF amod.next = NIL THEN BEGIN
          amod.desc.num := 1;
          amod.desc.next := 0;
        END ELSE BEGIN
          amod.desc.num := amod.next.desc.num + 1;
          amod.desc.next := amod.next.addr;
        END;
      END ELSE
        error(7, name1)
    END;

    IF res = 0 THEN BEGIN(*read file*)
      INC(p, _DIV(DescSize,4)); (*allocate descriptor*)
      (* mod.desc.name := name; *)
      n := 0; WHILE name[n] <> #0 DO BEGIN amod.desc.name[n] := name[n]; INC(n); END;
      amod.desc.key := key;
      amod.desc.refcnt := 0;
      amod.desc.data := p * 4;  (*data*)
      buffer[_DIV(MTOrg,4) + amod.desc.num] := p * 4;  (*module table entry*)
      ReadInt(F, n);
      WHILE n > 0 DO BEGIN
        ReadInt(F, w);
        buffer[p] := w;
        INC(p); DEC(n, 4);
      END;  (*type descriptors*)
      ReadInt(F, n);
      WHILE n > 0 DO BEGIN buffer[p] := 0; INC(p); DEC(n, 4) END;  (*variable space*)
      ReadInt(F, n);
      WHILE n > 0 DO BEGIN ReadInt(F, w); buffer[p] := w; INC(p); DEC(n, 4) END;  (*strings*)
      amod.desc.code := p * 4;  (*program*)
      ReadInt(F, n);
      WHILE n > 0 DO BEGIN ReadInt(F, w); buffer[p] := w; INC(p); DEC(n) END;  (*program code*)
      amod.desc.imp := p * 4;  (*copy imports*)
      i := 0;
      WHILE i < nofimps DO BEGIN
        buffer[p] := import[i].addr;
        INC(p);
        INC(i);
      END;
      amod.desc.cmd := p * 4;  (*commands*)
      ReadStringPart(F, w, n);
      WHILE n <> 0 DO BEGIN
        WHILE n = 4 DO BEGIN
          buffer[p] := w;
          INC(p);
          ReadStringPart(F, w, n);
        END;
        buffer[p] := w;
        INC(p);
        ReadInt(F, w);
        buffer[p] := w;
        INC(p);
        ReadStringPart(F, w, n);
      END;
      buffer[p] := 0;
      INC(p);
      amod.desc.ent := p * 4;  (*entries*)
      ReadInt(F, n);
      WHILE n > 0 DO BEGIN ReadInt(F, w); buffer[p] := w; INC(p); DEC(n) END;
      amod.desc.ptr := p * 4;  (*pointer references*)
      ReadInt(F, w);
      WHILE w >= 0 DO BEGIN
        buffer[p] := amod.desc.data + w;
        INC(p);
        ReadInt(F, w)
      END;
      buffer[p] := 0;
      INC(p);
      ReadInt(F, fixorgP);
      ReadInt(F, fixorgD);
      ReadInt(F, fixorgT);
      ReadInt(F, w);
      body := amod.desc.code + w;
      Read(F, ch);
      IF ch <> 'O' THEN BEGIN (*corrupted file*)  amod := NIL; error(4, name) END
    END;

    IF res = 0 THEN BEGIN (*fixup of BL*)
      adr := amod.desc.code + fixorgP*4;
      WHILE adr <> amod.desc.code DO BEGIN
        inst := buffer[_DIV(adr,4)];
        mno := _MOD(_DIV(inst,$100000),$10);
        pno := _MOD(_DIV(inst,$1000),$100);
        disp := _MOD(inst,$1000);
        impmod := import[mno-1];
        dest := buffer[_DIV(impmod.desc.ent,4) + pno] + impmod.desc.code;
        offset := _DIV((dest - adr - 4),4);
        buffer[_DIV(adr,4)] := _MOD(offset,$1000000) OR $0F7000000;
        adr := adr - disp*4
      END;
      (*fixup of LDR/STR/ADD*)
      adr := amod.desc.code + fixorgD*4;
      WHILE adr <> amod.desc.code DO BEGIN
        inst := buffer[_DIV(adr,4)];
        mno := _MOD(_DIV(inst,$100000),$10);
        disp := _MOD(inst,$1000);
        IF mno = 0 THEN (*global*)
          buffer[_DIV(adr,4)] := (_DIV(inst,$1000000) * $10 + MT) * $100000 + amod.desc.num * 4
        ELSE BEGIN (*import*)
          impmod := import[mno-1];
          v := impmod.desc.num;
          buffer[_DIV(adr,4)] := (_DIV(inst,$1000000) * $10 + MT) * $100000 + v*4;
          inst := buffer[_DIV(adr,4) + 1];
          vno := _MOD(inst,$100);
          offset := buffer[_DIV(impmod.desc.ent,4) + vno];
          IF ODD(_DIV(inst,$100)) THEN offset := offset + impmod.desc.code - impmod.desc.data;
          buffer[_DIV(adr,4) + 1] := _DIV(inst,$10000) * $10000 + offset;
        END;
        adr := adr - disp*4
      END;
      (*fixup of type descriptors*)
      adr := amod.desc.data + fixorgT*4;
      WHILE adr <> amod.desc.data DO BEGIN
        inst := buffer[_DIV(adr,4)];
        mno := _MOD(_DIV(inst,$1000000),$10);
        vno := _MOD(_DIV(inst,$1000),$1000);
        disp := _MOD(inst,$1000);
        ShowMessage(IntToStr(fixorgT)+' '+IntToStr(adr)+' '+IntToHex(inst,8)+' '+IntToHex(mno,8)+' '+IntToHex(vno,8)+' '+IntToHex(disp,8));
        IF mno = 0 THEN (*global*)
          inst := amod.desc.data + vno
        ELSE BEGIN (*import*)
          impmod := import[mno-1];
          offset := buffer[_DIV(impmod.desc.ent,4) + vno];
          inst := impmod.desc.data + offset
        END;
        buffer[_DIV(adr,4)] := inst;
        adr := adr - disp*4;
      END ;
      buffer[0] := $0E7000000 OR _DIV(body,4) - 1   (*initialize module*)
    END ELSE IF res = 3 THEN BEGIN
      importing := name;
      WHILE nofimps > 0 DO BEGIN
        DEC(nofimps);
        DEC(import[nofimps].desc.refcnt)
      END;
    END;
    F.free;
  END;

  newmod := amod;

END;

PROCEDURE Link(path,name: ARRAY OF CHAR; VAR buffer: ARRAY OF INTEGER);
VAR
  amod: Module;
  n: LONGINT;
BEGIN
  FOR n := 0 TO LENGTH(buffer) - 1 DO buffer[n] := 0;
  AllocPtr := $100;
  Load(path,name, buffer, amod);
  buffer[4] := AllocPtr;
  buffer[5] := amod.addr;
  buffer[6] := $40000;  (*module limit*)
  (*store module descriptors*)
  WHILE root <> NIL DO BEGIN
    MOVE(root.desc, buffer[_DIV(root.addr,4)],SizeOf(root.desc));  // _DIV(DescSize,4)
    root := root.next;
  END;
END;

PROCEDURE LinkDiskImage(path,modname: ARRAY OF CHAR; corename: ARRAY OF CHAR);
VAR
  buff: Buffer;
  F : TFileStream;
  i : INTEGER;
  fpath : Base.TStrings;
BEGIN
  GC := TObjectList.Create;
  fpath := path+Base.objdir;
  Link(fpath,modname, buff);
  IF res = 0 THEN BEGIN
    F := TFileStream.Create(path+corename+Base.binext,fmCreate);
    F.Seek(0,soFromBeginning);
    i := $9B1EA38D;
    F.Write(i,4);
    FOR i := 4 TO 1023 DO F.WriteByte(0);
    FOR i := 0 TO _DIV(AllocPtr,4) - 1 DO F.Write(buff[i],4);
    F.free;
  END;
  GC.free;
END;

END.

