UNIT Base;

{$mode objfpc}{$H+}

INTERFACE

USES
  Classes, Contnrs, SysUtils, Dialogs;

CONST
  symext     = '.smb';
  objext     = '.rsc';
  binext     = '.bin';
  objdir     = 'obj'+ DirectorySeparator;

  WordSize   = 4;
  IdLen      = 32;
  BufSize    = 256;
  versionkey = 1;
  maxTypTab  = 64;

  cHead = 0; cConst = 1; cVar   = 2; cPar   = 3;
  cFld  = 4; cTyp   = 5; cSProc = 6; cSFunc = 7; cMod = 8;

  fByte   = 1; fBool = 2;  fChar    = 3;  fInt    = 4;
  fReal   = 5; fSet  = 6;  fPointer = 7;  fNilTyp = 8;
  fNoTyp  = 9; fProc = 10; fString  = 11; fArray  = 12;
  fRecord = 13;

TYPE
  TIdent   = ARRAY [0..IdLen-1] OF CHAR;
  TStrings = ARRAY [0..BufSize-1] OF CHAR;

  Object_ = CLASS;
  Module_ = CLASS;
  Type_   = CLASS;

  Object_ = CLASS
    clas   : INTEGER;
    lev    : INTEGER;
    exno   : INTEGER;
    expo   : BOOLEAN;
    rdo    : BOOLEAN;
    next   : Object_;
    dsc    : Object_;
    typ    : Type_;
    name   : TIdent;
    val    : LONGINT;
  END;

  Module_ = CLASS(Object_)
    orgname: TIdent;
  END ;

  Type_ = CLASS
    form   : INTEGER;
    ref    : INTEGER;
    mno    : INTEGER;
    nofpar : INTEGER;
    len    : LONGINT;
    dsc    : Object_;
    typobj : Object_;
    base   : Type_;
    size   : LONGINT;
  END;

  Item = RECORD
    mode : INTEGER;
    typ  : Type_;
    a,b,r: LONGINT;
    rdo  : BOOLEAN;
  END;

  ProcClear = PROCEDURE;

VAR
  objlist: TObjectList;
  topScope, universe, vsystem : Object_;
  byteType, boolType, charType: Type_;
  intType, realType, setType  : Type_;
  nilType, noType, strType    : Type_;
  nofmod, ref: INTEGER;
  typtab: ARRAY [0..maxTypTab-1] OF Type_;
  dummy : Object_;

  errcnt : INTEGER;
  errpos : LONGINT;
  filepos: INTEGER;
  charerr    : INTEGER;
  lineerr    : INTEGER;
  current    : TStrings;

  FUNCTION FLT(CONST I: INTEGER): SINGLE;
  FUNCTION _DIV(CONST x, y: INTEGER): INTEGER;
  FUNCTION _MOD(CONST x, y: INTEGER): INTEGER;
  FUNCTION LSL(CONST x, y: INTEGER): INTEGER;
  FUNCTION ASR(CONST x, y: INTEGER): INTEGER;
  FUNCTION ROR(CONST x, y: INTEGER): INTEGER;
  FUNCTION HASH(P: POINTER; L: INTEGER): INTEGER;

  PROCEDURE Mark(msg: ARRAY OF CHAR);
  PROCEDURE Init;
  PROCEDURE Reset;
  PROCEDURE OpenScope;
  PROCEDURE CloseScope;

  PROCEDURE NewObj(VAR obj: Object_; id: TIdent; clas: INTEGER);
  PROCEDURE NewType(VAR tp: Type_; form: INTEGER; size: LONGINT);
  FUNCTION  ThisObj(id: TIdent): Object_;
  FUNCTION  ThisImport(id: TIdent; amod: Object_): Object_;
  FUNCTION  ThisField(id: TIdent; rec: Type_): Object_;
  FUNCTION  ThisModule(name, orgname: TIdent; non: BOOLEAN; key: LONGINT): Object_;

  PROCEDURE MakeConstItem(VAR x: Item; typ: Type_; val: LONGINT);
  PROCEDURE MakeRealItem(VAR x: Item; val: SINGLE);
  PROCEDURE MakeItem(VAR x: Item; y: Object_; curlev: LONGINT);

  PROCEDURE InitSystem(VAR modid, modid1: TIdent);

IMPLEMENTATION

USES
  Main;

FUNCTION FLT(CONST I: INTEGER): SINGLE;
BEGIN
  RESULT := I*1.0;
END;

FUNCTION _DIV(CONST x, y: INTEGER): INTEGER;
VAR
  Q,R: INTEGER;
BEGIN
  Q:=TRUNC(x/y);
  R:=x MOD y;
  IF R<0 THEN BEGIN
    IF y>0 THEN
      Q:=Q-1
    ELSE
      Q:=Q+1;
  END;
  RESULT := Q;
END;

FUNCTION _MOD(CONST x, y: INTEGER): INTEGER;
BEGIN
  ASSERT(y > 0);
  IF x >= 0 THEN
    RESULT := x MOD y
  ELSE
    RESULT := (y + (x MOD y)) MOD y;
END;

FUNCTION LSL(CONST x, y: INTEGER): INTEGER;
BEGIN
  RESULT := x SHL (y AND 31);
END;

FUNCTION ASR(CONST x, y: INTEGER): INTEGER;
BEGIN
  RESULT := (LONGWORD(x)) SHR (y AND 31);
END;

FUNCTION ROR(CONST x, y: INTEGER): INTEGER;
BEGIN
  RESULT := (x SHR (y AND 31)) OR (x SHL (-y AND 31));
END;

FUNCTION INTTOREAL(CONST I: INTEGER): SINGLE;
VAR
  R: SINGLE ABSOLUTE I;
BEGIN
  RESULT := R;
END;

FUNCTION REALTOINT(CONST R: SINGLE): INTEGER;
VAR
  I: INTEGER ABSOLUTE R;
BEGIN
  RESULT := I;
END;

FUNCTION HASH(P: POINTER; L: INTEGER): INTEGER;
VAR
  I: POINTER;
  x: INTEGER;
BEGIN
  I := P;
  RESULT := 0;
  x := 0;
  WHILE x < L DIV 2 DO BEGIN
    RESULT := RESULT * 1664525 + BYTE(I^) + 1013904223;
    INC(I,2);
    INC(x);
  END;
END;

PROCEDURE Mark(msg: ARRAY OF CHAR);
VAR
  p: LONGINT;
BEGIN
  p := filepos;

  IF errcnt = 0 THEN BEGIN
    Form1.Document.OpenPageByName(current);
    Form1.pagecontrol.InvisibleTab(0);
    Form1.document.Editor.ErrorLine := lineerr+1;
    Form1.document.Editor.CaretXY := Point(charerr-1,lineerr+1);
    Form1.document.Editor.SetFocus;
    Form1.document.Editor.SynMarkup.enabled := false;
    Form1.document.UpdateText;
    Form1.pagecontrol.Invalidate;
    Form1.document.Editor.TopLine  := Form1.document.Editor.CaretY-(Form1.document.Editor.LinesInWindow div 4);
  END;

  IF (p > errpos) AND (errcnt < 25) THEN BEGIN
    Form1.msgview.AddMessage(current+' '+'['+IntToStr(lineerr+1)+':'+IntToStr(charerr-1)+'] '+msg);
  END;

  INC(errcnt);
  errpos := p + 4;
END;

FUNCTION ThisObj(id: TIdent): Object_;
VAR
  s, x: Object_;
BEGIN
  s := topScope;
  REPEAT
    x := s.next;
    WHILE (x <> NIL) AND (x.name <> id) DO x := x.next;
    s := s.dsc;
  UNTIL (x <> NIL) OR (s = NIL);
  RESULT := x;
END;

FUNCTION ThisImport(id: TIdent; amod: Object_): Object_;
VAR
  obj: Object_;
BEGIN
  IF amod.rdo THEN BEGIN
    IF amod.name[0] <> #0 THEN BEGIN
      obj := amod.dsc;
      WHILE (obj <> NIL) AND (obj.name <> id) DO obj := obj.next;
    END ELSE
      obj := NIL;
  END ELSE
    obj := NIL;
  RESULT := obj
END;

FUNCTION ThisField(id: TIdent; rec: Type_): Object_;
VAR
  fld: Object_;
BEGIN
  fld := rec.dsc;
  WHILE (fld <> NIL) AND (fld.name <> id) DO fld := fld.next;
  RESULT := fld;
END;

FUNCTION ThisModule(name, orgname: TIdent; non: BOOLEAN; key: LONGINT): Object_;
VAR
  newmod: Module_;
  obj, obj1: Object_;
BEGIN

  obj1 := topScope;
  obj := obj1.next;

  WHILE (obj <> NIL) AND (obj.name <> name) DO BEGIN
    obj1 := obj;
    obj := obj1.next
  END;

  IF obj = NIL THEN BEGIN
    newmod := Module_.Create;
    newmod.clas    := cMod;
    newmod.rdo     := FALSE;
    newmod.name    := COPY(name,0,length(name));
    newmod.orgname := orgname;
    newmod.val     := key;
    newmod.lev     := nofmod;
    newmod.typ     := noType;
    newmod.dsc     := NIL;
    newmod.next    := NIL;
    INC(nofmod);
    obj1.next      := newmod;
    obj            := newmod;
    objlist.Add(newmod);
  END ELSE BEGIN
    IF non THEN Mark('invalid import order');
  END;

  RESULT := obj;

END;

PROCEDURE NewObj(VAR obj: Object_; id: TIdent; clas: INTEGER);
VAR
  anew, x: Object_;
BEGIN
  x := topScope;
  WHILE (x.next <> NIL) AND (x.next.name <> id) DO x := x.next;
  IF x.next = NIL THEN BEGIN
    anew      := Object_.Create;
    anew.name := id;
    anew.clas := clas;
    anew.next := NIL;
    anew.rdo  := FALSE;
    anew.dsc  := NIL;
    x.next    := anew;
    obj       := anew;
    objlist.Add(anew);
  END ELSE BEGIN
    obj := x.next;
    Mark('mult def');
  END;
END;

PROCEDURE NewType(VAR tp: Type_; form: INTEGER; size: LONGINT);
BEGIN
  tp := Type_.Create;
  tp.form := form;
  tp.size := size;
  tp.ref  := 0;
  tp.base := NIL;
  objlist.Add(tp);
END;

PROCEDURE StdType(VAR tp: Type_; ref, form: INTEGER; size: LONGINT);
BEGIN
  NewType(tp,form,size);
  tp.ref  := ref;
  tp.base := NIL;
  typtab[ref] := tp;
END;

PROCEDURE MakeConstItem(VAR x: Item; typ: Type_; val: LONGINT);
BEGIN
  x.mode := cConst;
  x.typ := typ;
  x.a := val;
END;

PROCEDURE MakeRealItem(VAR x: Item; val: SINGLE);
BEGIN
  x.mode := cConst;
  x.typ  := realType;
  x.a    := REALTOINT(val);
END;

PROCEDURE MakeItem(VAR x: Item; y: Object_; curlev: LONGINT);
BEGIN
  x.mode := y.clas;
  x.typ  := y.typ;
  x.a    := y.val;
  x.rdo  := y.rdo;
  IF y.clas = cPar THEN
    x.b := 0
  ELSE IF y.clas = cTyp THEN BEGIN
    x.a := y.typ.len;
    x.r := -y.lev;
  END ELSE IF (y.clas = cConst) AND (y.typ.form = fString) THEN
    x.b := y.lev
  ELSE
    x.r := y.lev;
  IF (y.lev > 0) AND (y.lev <> curlev) AND (y.clas <> cConst) THEN
    Mark('level error, not accessible');
END;

PROCEDURE Enter(name: ARRAY OF CHAR; cl: INTEGER; typ: Type_; n: LONGINT);
VAR
  obj: Object_;
BEGIN
  obj := Object_.Create;
  obj.name := COPY(name,0,length(name));
  obj.clas := cl;
  obj.typ  := typ;
  obj.val  := n;
  obj.dsc  := NIL;
  IF cl = cTyp THEN typ.typobj := obj;
  obj.next := vsystem;
  vsystem  := obj;
  objlist.Add(obj);
END;

PROCEDURE OpenScope;
VAR
  s: Object_;
BEGIN
  s := Object_.Create;
  s.clas := cHead;
  s.dsc  := topScope;
  s.next := NIL;
  topScope := s;
  objlist.Add(s);
END;

PROCEDURE CloseScope;
BEGIN
  topScope := topScope.dsc
END;

PROCEDURE InitSystem(VAR modid, modid1: TIdent);
VAR
  key: LONGINT;
  thismod: Object_;
BEGIN
  key := 0;
  thismod := ThisModule(modid, modid1,TRUE,key);
  DEC(nofmod);
  thismod.lev := 0;
  thismod.dsc := vsystem;
  thismod.rdo := TRUE;
END;

PROCEDURE Init;
BEGIN

  current := '';

  topScope := universe;
  nofmod   := 1;

  objlist := TObjectList.Create;

  vsystem := NIL;

  StdType(byteType,fByte  ,fInt   , 1);
  StdType(boolType,fBool  ,fBool  , 1);
  StdType(charType,fChar  ,fChar  , 1);
  StdType(intType ,fInt   ,fInt   , 4);
  StdType(realType,fReal  ,fReal  , 4);
  StdType(setType ,fSet   ,fSet   , 4);
  StdType(nilType ,fNilTyp,fNilTyp, 4);
  StdType(noType  ,fNoTyp ,fNoTyp , 4);
  StdType(strType ,fString,fString, 8);

  Enter('UML'   , cSFunc, intType , 132);
  Enter('SBC'   , cSFunc, intType , 122);
  Enter('ADC'   , cSFunc, intType , 112);
  Enter('ROR'   , cSFunc, intType , 92);
  Enter('ASR'   , cSFunc, intType , 82);
  Enter('LSL'   , cSFunc, intType , 72);
  Enter('LEN'   , cSFunc, intType , 61);
  Enter('CHR'   , cSFunc, charType, 51);
  Enter('ORD'   , cSFunc, intType , 41);
  Enter('FLT'   , cSFunc, realType, 31);
  Enter('FLOOR' , cSFunc, intType , 21);
  Enter('ODD'   , cSFunc, boolType, 11);
  Enter('ABS'   , cSFunc, intType , 1);

  Enter('LED'   , cSProc, noType  , 81);
  Enter('UNPK'  , cSProc, noType  , 72);
  Enter('PACK'  , cSProc, noType  , 62);
  Enter('NEW'   , cSProc, noType  , 51);
  Enter('ASSERT', cSProc, noType  , 41);
  Enter('EXCL'  , cSProc, noType  , 32);
  Enter('INCL'  , cSProc, noType  , 22);
  Enter('DEC'   , cSProc, noType  , 11);
  Enter('INC'   , cSProc, noType  , 1);

  Enter('SET'     , cTyp, setType , 0);
  Enter('BYTE'    , cTyp, byteType, 0);
  Enter('CHAR'    , cTyp, charType, 0);
  Enter('REAL'    , cTyp, realType, 0);
  Enter('BOOLEAN' , cTyp, boolType, 0);
  Enter('LONGINT' , cTyp, intType , 0);
  Enter('INTEGER' , cTyp, intType , 0);
  Enter('LONGREAL', cTyp, realType, 0);

  topScope := NIL;
  OpenScope;
  topScope.next := vsystem;
  universe := topScope;
  vsystem := NIL;

  Enter('H'     , cSFunc, intType , 201);
  Enter('COND'  , cSFunc, boolType, 191);
  Enter('SIZE'  , cSFunc, intType , 181);
  Enter('ADR'   , cSFunc, intType , 171);
  Enter('VAL'   , cSFunc, intType , 162);
  Enter('REG'   , cSFunc, intType , 151);
  Enter('BIT'   , cSFunc, boolType, 142);
  Enter('LDREG' , cSProc, noType  , 142);
  Enter('LDPSR' , cSProc, noType  , 131);
  Enter('COPY'  , cSProc, noType  , 123);
  Enter('PUT'   , cSProc, noType  , 112);
  Enter('GET'   , cSProc, noType  , 102);

  dummy := Object_.Create;
  dummy.clas := cVar;
  dummy.typ  := intType;
  objlist.Add(dummy);

END;

PROCEDURE Reset;
BEGIN
  FOR Ref := fRecord+1 TO maxTypTab-1 DO BEGIN
    typtab[Ref] := NIL
  END;
  objlist.free;
END;

END.

