UNIT Expr;

{$mode objfpc}{$H+}

INTERFACE

USES
  Base, Scan, Gen, Func;

TYPE
  PtrBase = CLASS
    name: Base.TIdent;
    typ : Base.Type_;
    row : INTEGER;
    col : INTEGER;
    next: PtrBase;
  END;

VAR
  sym: INTEGER;
  dc : LONGINT;
  level, exno, version: INTEGER;
  pbsList: PtrBase;

  PROCEDURE Check(s: INTEGER; msg: ARRAY OF CHAR);
  PROCEDURE CheckBool(VAR x: Base.Item);
  PROCEDURE CheckInt(VAR x: Base.Item);
  PROCEDURE CheckReal(VAR x: Base.Item);
  PROCEDURE CheckSet(VAR x: Base.Item);
  PROCEDURE CheckSetVal(VAR x: Base.Item);
  PROCEDURE CheckConst(VAR x: Base.Item);
  PROCEDURE CheckReadOnly(VAR x: Base.Item);
  PROCEDURE CheckExport(VAR expo: BOOLEAN);
  FUNCTION  IsExtension(t0, t1: Base.Type_): BOOLEAN;
  PROCEDURE CheckRecLevel(lev: INTEGER);
  FUNCTION  CompTypes(t0, t1: Base.Type_; varpar: BOOLEAN): BOOLEAN;

  PROCEDURE ParamList(VAR x: Base.Item);
  PROCEDURE TypeTest(VAR x: Base.Item; T: Base.Type_; guard: BOOLEAN);
  PROCEDURE Qualident(VAR obj: Base.Object_);
  PROCEDURE Selector(VAR x: Base.Item);
  PROCEDURE Expression(VAR x: Base.Item);

IMPLEMENTATION

PROCEDURE Check(s: INTEGER; msg: ARRAY OF CHAR);
BEGIN
  IF sym = s THEN Scan.Get(sym) ELSE Base.Mark(msg);
END;

PROCEDURE CheckBool(VAR x: Base.Item);
BEGIN
  IF x.typ.form <> Base.fBool THEN BEGIN
    Base.Mark('not Boolean');
    x.typ := Base.boolType
  END;
END;

PROCEDURE CheckInt(VAR x: Base.Item);
BEGIN
  IF x.typ.form <> Base.fInt THEN BEGIN
    Base.Mark('not Integer');
    x.typ := Base.intType
  END;
END;

PROCEDURE CheckReal(VAR x: Base.Item);
BEGIN
  IF x.typ.form <> Base.fReal THEN BEGIN
    Base.Mark('not Real');
    x.typ := Base.realType
  END;
END;

PROCEDURE CheckSet(VAR x: Base.Item);
BEGIN
  IF x.typ.form <> Base.fSet THEN BEGIN
    Base.Mark('not Set');
    x.typ := Base.setType
  END;
END;

PROCEDURE CheckSetVal(VAR x: Base.Item);
BEGIN
  IF x.typ.form <> Base.fInt THEN BEGIN
    Base.Mark('not Int');
    x.typ := Base.setType;
  END ELSE IF x.mode = Base.cConst THEN BEGIN
    IF (x.a < 0) OR (x.a >= 32) THEN Base.Mark('invalid set');
  END
END;

PROCEDURE CheckConst(VAR x: Base.Item);
BEGIN
  IF x.mode <> Base.cConst THEN BEGIN
    Base.Mark('not a constant');
    x.mode := Base.cConst;
  END;
END;

PROCEDURE CheckReadOnly(VAR x: Base.Item);
BEGIN
  IF x.rdo THEN Base.Mark('read-only');
END;

PROCEDURE CheckExport(VAR expo: BOOLEAN);
BEGIN
  IF sym = Scan.lx_times THEN BEGIN
    expo := TRUE;
    Scan.Get(sym);
    IF level <> 0 THEN Base.Mark('remove asterisk');
  END ELSE
    expo := FALSE;
END;

FUNCTION IsExtension(t0, t1: Base.Type_): BOOLEAN;
BEGIN
  RESULT := (t0 = t1) OR (t1 <> NIL) AND IsExtension(t0, t1.base)
END;

PROCEDURE CheckRecLevel(lev: INTEGER);
BEGIN
  IF lev <> 0 THEN Base.Mark('ptr base must be global');
END;

FUNCTION CompTypes(t0, t1: Base.Type_; varpar: BOOLEAN): BOOLEAN;

  FUNCTION EqualSignatures(t0, t1: Base.Type_): BOOLEAN;
  VAR
    p0, p1: Base.Object_;
    com: BOOLEAN;
  BEGIN
    com := TRUE;
    IF (t0.base = t1.base) AND (t0.nofpar = t1.nofpar) THEN BEGIN
      p0 := t0.dsc;
      p1 := t1.dsc;
      WHILE p0 <> NIL DO BEGIN
        IF (p0.clas = p1.clas) AND  (p0.rdo = p1.rdo) AND
          ((p0.typ = p1.typ) OR
          (p0.typ.form = Base.fArray) AND (p1.typ.form = Base.fArray) AND (p0.typ.len = p1.typ.len) AND (p0.typ.base = p1.typ.base) OR
          (p0.typ.form = Base.fProc) AND (p1.typ.form = Base.fProc) AND EqualSignatures(p0.typ, p1.typ))
        THEN BEGIN
          p0 := p0.next;
          p1 := p1.next;
        END ELSE BEGIN
          p0 := NIL;
          com := FALSE ;
        END;
      END;
    END ELSE
      com := FALSE;
    RESULT := com;
  END;

BEGIN
  RESULT := (t0 = t1)
    OR (t0.form = Base.fArray) AND (t1.form = Base.fArray) AND (t0.base =  t1.base) AND (t0.len = t1.len)
    OR (t0.form = Base.fRecord) AND (t1.form = Base.fRecord)  AND IsExtension(t0, t1)
    OR NOT varpar AND
      ((t0.form = Base.fPointer) AND (t1.form = Base.fPointer)  AND IsExtension(t0.base, t1.base)
      OR (t0.form = Base.fProc) AND (t1.form = Base.fProc) AND EqualSignatures(t0, t1)
      OR (t0.form IN [Base.fPointer, Base.fProc]) AND (t1.form = Base.fNilTyp))
END;

(*********************************)

PROCEDURE Parameter(par: Base.Object_);
VAR
  x: Base.Item;
  varpar: BOOLEAN;
BEGIN

  Expression(x);

  IF par <> NIL THEN BEGIN

    varpar := par.clas = Base.cPar;
    IF CompTypes(par.typ, x.typ, varpar) THEN BEGIN
      IF NOT varpar THEN
        Gen.ValueParam(x)
      ELSE BEGIN
        IF NOT par.rdo THEN CheckReadOnly(x);
        Gen.VarParam(x, par.typ)
      END;
    END ELSE IF (x.typ.form = Base.fArray) AND (par.typ.form = Base.fArray)
    AND (x.typ.base = par.typ.base) AND (par.typ.len < 0) THEN BEGIN
      IF NOT par.rdo THEN CheckReadOnly(x);
      Gen.OpenArrayParam(x)
    END ELSE IF (x.typ.form = Base.fString) AND varpar AND par.rdo AND (par.typ.form = Base.fArray)
    AND (par.typ.base.form = Base.fChar) AND (par.typ.len < 0) THEN
      Gen.StringParam(x)
    ELSE IF NOT varpar AND (par.typ.form = Base.fInt) AND (x.typ.form = Base.fInt) THEN
      Gen.ValueParam(x)
    ELSE IF (x.typ.form = Base.fString) AND (x.b = 2) AND (par.clas = Base.cVar) AND (par.typ.form = Base.fChar) THEN BEGIN
      Gen.StrToChar(x);
      Gen.ValueParam(x)
    END ELSE IF (par.typ.form = Base.fArray) AND (par.typ.base = Base.byteType)
    AND (par.typ.len > 0) AND (par.typ.size = x.typ.size) THEN
      Gen.VarParam(x, par.typ)
    ELSE
      Base.Mark('incompatible parameters');

  END;

END;

PROCEDURE ParamList(VAR x: Base.Item);
VAR
  n: INTEGER;
  par: Base.Object_;
BEGIN
  par := x.typ.dsc;
  n := 0;

  IF sym <> Scan.lx_rparen THEN BEGIN
    Parameter(par);
    n := 1;
    WHILE sym <= Scan.lx_comma DO BEGIN
      Check(sym, 'comma?');
      IF par <> NIL THEN par := par.next;
      INC(n);
      Parameter(par);
    END;
    Check(Scan.lx_rparen, ') missing');
  END ELSE
    Scan.Get(sym);

  IF n < x.typ.nofpar THEN
    Base.Mark('too few params')
  ELSE IF n > x.typ.nofpar THEN
    Base.Mark('too many params');

END;

PROCEDURE TypeTest(VAR x: Base.Item; T: Base.Type_; guard: BOOLEAN);
VAR
  xt: Base.Type_;
BEGIN
  xt := x.typ;
  IF (T.form IN [Base.fPointer, Base.fRecord]) AND (T.form = xt.form) THEN BEGIN
    WHILE (xt <> T) AND (xt <> NIL) DO xt := xt.base;
    IF xt <> T THEN BEGIN
      xt := x.typ;
      IF xt.form = Base.fPointer THEN BEGIN
        IF IsExtension(xt.base, T.base) THEN BEGIN
          Gen.TypeTest(x, T.base, FALSE, guard);
          x.typ := T;
        END ELSE
          Base.Mark('not an extension');
      END ELSE IF (xt.form = Base.fRecord) AND (x.mode = Base.cPar) THEN BEGIN
        IF IsExtension(xt, T) THEN  BEGIN
          Gen.TypeTest(x, T, TRUE, guard);
          x.typ := T;
        END ELSE
          Base.Mark('not an extension');
      END ELSE
        Base.Mark('incompatible types');
    END ELSE IF NOT guard THEN
      Base.MakeConstItem(x, Base.boolType, 1);
  END ELSE
    Base.Mark('type mismatch');;
  IF NOT guard THEN x.typ := Base.boolType;
END;

PROCEDURE Qualident(VAR obj: Base.Object_);
BEGIN
  obj := Base.ThisObj(Scan.id);
  Scan.Get(sym);
  IF obj = NIL THEN BEGIN
    Base.Mark('undef');
    obj := Base.dummy;
  END;
  IF (sym = Scan.lx_period) AND (obj.clas = Base.cMod) THEN BEGIN
    Scan.Get(sym);
    IF sym = Scan.lx_ident THEN BEGIN
      obj := Base.ThisImport(Scan.id,obj);
      Scan.Get(sym);
      IF obj = NIL THEN BEGIN
        Base.Mark('undef');
        obj := dummy;
      END;
    END ELSE BEGIN
      Base.Mark('identifier expected');
      obj := dummy;
    END;
  END;
END;

PROCEDURE Selector(VAR x: Base.Item);
VAR
  y: Base.Item;
  obj: Base.Object_;
BEGIN
  WHILE (sym = Scan.lx_lbrak) OR (sym = Scan.lx_period) OR (sym = Scan.lx_arrow)
  OR (sym = Scan.lx_lparen) AND (x.typ.form IN [Base.fRecord, Base.fPointer]) DO BEGIN

    IF sym = Scan.lx_lbrak THEN BEGIN

      REPEAT
        Scan.Get(sym);
        Expression(y);
        IF x.typ.form = Base.fArray THEN BEGIN
          CheckInt(y);
          Gen.Index(x, y);
          x.typ := x.typ.base;
        END ELSE
          Base.Mark('not an array');
      UNTIL sym <> Scan.lx_comma;
      Check(Scan.lx_rbrak, 'no ]');

    END ELSE IF sym = Scan.lx_period THEN BEGIN

      Scan.Get(sym);
      IF sym = Scan.lx_ident THEN BEGIN
        IF x.typ.form = Base.fPointer THEN BEGIN
          Gen.DeRef(x);
          x.typ := x.typ.base;
        END;
        IF x.typ.form = Base.fRecord THEN BEGIN
          obj := Base.ThisField(Scan.id,x.typ);
          Scan.Get(sym);
          IF obj <> NIL THEN BEGIN
            Gen.Field(x, obj);
            x.typ := obj.typ;
          END ELSE
            Base.Mark('undef');
        END ELSE
          Base.Mark('not a record');
      END ELSE
        Base.Mark('ident?');

    END ELSE IF sym = Scan.lx_arrow THEN BEGIN
      Scan.Get(sym);
      IF x.typ.form = Base.fPointer THEN BEGIN
        Gen.DeRef(x);
        x.typ := x.typ.base;
      END ELSE
        Base.Mark('not a pointer');
    END ELSE IF (sym = Scan.lx_lparen) AND (x.typ.form IN [Base.fRecord, Base.fPointer]) THEN BEGIN
      Scan.Get(sym);
      IF sym = Scan.lx_ident THEN BEGIN
        Qualident(obj);
        IF obj.clas = Base.cTyp THEN
          TypeTest(x, obj.typ, TRUE)
        ELSE
          Base.Mark('guard type expected');
      END ELSE
        Base.Mark('not an identifier');
      Check(Scan.lx_rparen, ' ) missing');
    END;

  END;
END;

PROCEDURE Element(VAR x: Base.Item);
VAR
  y: Base.Item;
BEGIN
  Expression(x);
  CheckSetVal(x);
  IF sym = Scan.lx_upto THEN BEGIN
    Scan.Get(sym);
    Expression(y);
    CheckSetVal(y);
    Gen.Set2(x, y)
  END ELSE
    Gen.Singleton(x);
  x.typ := Base.setType;
END;

PROCEDURE Sets(VAR x: Base.Item);
VAR
  y: Base.Item;
BEGIN
  IF sym >= Scan.lx_if THEN BEGIN
    IF sym <> Scan.lx_rbrace THEN Base.Mark(' } missing');
    Base.MakeConstItem(x, Base.setType, 0);
  END ELSE BEGIN
    Element(x);
    WHILE (sym < Scan.lx_rparen) OR (sym > Scan.lx_rbrace) DO BEGIN
      IF sym = Scan.lx_comma THEN
        Scan.Get(sym)
      ELSE IF sym <> Scan.lx_rbrace THEN
        Base.Mark('missing comma');
      Element(y);
      Gen.SetOp(Scan.lx_plus, x, y)
    END;
  END;
END;

PROCEDURE Factor(VAR x: Base.Item);
VAR
  obj: Base.Object_;
  rx: LONGINT;
BEGIN

  IF (sym < Scan.lx_char) OR (sym > Scan.lx_ident) THEN BEGIN
    Base.Mark('expression expected');
    REPEAT
      Scan.Get(sym)
    UNTIL (sym >= Scan.lx_char) AND (sym <= Scan.lx_ident);
  END ;

  IF sym = Scan.lx_ident THEN BEGIN
    Qualident(obj);
    IF obj.clas = Base.cSFunc THEN
      Func.StandFunc(x, obj.val, obj.typ)
    ELSE BEGIN
      Base.MakeItem(x, obj, level);
      Selector(x);
      IF sym = Scan.lx_lparen THEN BEGIN
        Scan.Get(sym);
        IF (x.typ.form = Base.fProc) AND (x.typ.base.form <> Base.fNoTyp) THEN BEGIN
          Gen.PrepCall(x, rx);
          ParamList(x);
          Gen.Call(x, rx);
          x.typ := x.typ.base;
        END ELSE BEGIN
          Base.Mark('not a function');
          ParamList(x);
        END;
      END;
    END;
  END ELSE IF sym = Scan.lx_int    THEN BEGIN Base.MakeConstItem(x, Base.intType,  Scan.ival);  Scan.Get(sym);
  END ELSE IF sym = Scan.lx_real   THEN BEGIN Base.MakeRealItem (x, Scan.rval);                 Scan.Get(sym);
  END ELSE IF sym = Scan.lx_char   THEN BEGIN Base.MakeConstItem(x, Base.charType, Scan.ival);  Scan.Get(sym);
  END ELSE IF sym = Scan.lx_nil    THEN BEGIN Base.MakeConstItem(x, Base.nilType, 0);           Scan.Get(sym);
  END ELSE IF sym = Scan.lx_false  THEN BEGIN Base.MakeConstItem(x, Base.boolType, 0);          Scan.Get(sym);
  END ELSE IF sym = Scan.lx_true   THEN BEGIN Base.MakeConstItem(x, Base.boolType, 1);          Scan.Get(sym);
  END ELSE IF sym = Scan.lx_string THEN BEGIN Gen.MakeStringItem(x, Scan.slen);                 Scan.Get(sym);
  END ELSE IF sym = Scan.lx_lparen THEN BEGIN Scan.Get(sym); Expression(x); Check(Scan.lx_rparen, 'no )')
  END ELSE IF sym = Scan.lx_lbrace THEN BEGIN Scan.Get(sym); Sets(x);       Check(Scan.lx_rbrace, 'no }')
  END ELSE IF sym = Scan.lx_not    THEN BEGIN Scan.Get(sym); Factor(x);     CheckBool(x); Gen.Not1(x)
  END ELSE BEGIN
    Base.Mark('not a factor');
    Base.MakeConstItem(x, Base.intType, 0);
  END;

END;

PROCEDURE Term(VAR x: Base.Item);
VAR
  y: Base.Item;
  op, f: INTEGER;
BEGIN
  Factor(x);

  IF x.typ = NIL THEN x.typ := Base.noType;

  f := x.typ.form;

  WHILE (sym >= Scan.lx_times) AND (sym <= Scan.lx_and) DO BEGIN
    op := sym;
    Scan.Get(sym);
    IF op = Scan.lx_times THEN BEGIN
      IF          f = Base.fInt  THEN BEGIN Factor(y); CheckInt(y);  Gen.MulOp(x, y)
      END ELSE IF f = Base.fReal THEN BEGIN Factor(y); CheckReal(y); Gen.RealOp(op, x, y)
      END ELSE IF f = Base.fSet  THEN BEGIN Factor(y); CheckSet(y);  Gen.SetOp(op, x, y)
      END ELSE
        Base.Mark('bad type');
    END ELSE IF (op = Scan.lx_div) OR (op = Scan.lx_mod) THEN BEGIN
      CheckInt(x);
      Factor(y);
      CheckInt(y);
      Gen.DivOp(op, x, y)
    END ELSE IF op = Scan.lx_rdiv THEN BEGIN
      IF          f = Base.fReal THEN BEGIN Factor(y); CheckReal(y); Gen.RealOp(op, x, y)
      END ELSE IF f = Base.fSet  THEN BEGIN Factor(y); CheckSet(y);  Gen.SetOp(op, x, y)
      END ELSE
        Base.Mark('bad type');
    END ELSE BEGIN
      CheckBool(x);
      Gen.And1(x);
      Factor(y);
      CheckBool(y);
      Gen.And2(x, y)
    END;
  END;
END;

PROCEDURE SimpleExpression(VAR x: Base.Item);
VAR
  y: Base.Item;
  op: INTEGER;
BEGIN

  IF sym = Scan.lx_minus THEN BEGIN
    Scan.Get(sym);
    Term(x);
    IF x.typ.form IN [Base.fInt, Base.fReal, Base.fSet] THEN
      Gen.Neg(x)
    ELSE
      CheckInt(x);
  END ELSE IF sym = Scan.lx_plus THEN BEGIN
    Scan.Get(sym);
    Term(x);
  END ELSE
    Term(x);

  WHILE (sym >= Scan.lx_plus) AND (sym <= Scan.lx_or) DO BEGIN
    op := sym;
    Scan.Get(sym);
    IF op = Scan.lx_or THEN BEGIN
      Gen.Or1(x);
      CheckBool(x);
      Term(y);
      CheckBool(y);
      Gen.Or2(x, y)
    END ELSE IF x.typ.form = Base.fInt THEN BEGIN
      Term(y);
      CheckInt(y);
      Gen.AddOp(op, x, y)
    END ELSE IF x.typ.form = Base.fReal THEN BEGIN
      Term(y);
      CheckReal(y);
      Gen.RealOp(op, x, y)
    END ELSE BEGIN
      CheckSet(x);
      Term(y);
      CheckSet(y);
      Gen.SetOp(op, x, y)
    END;
  END;

END;

PROCEDURE Expression(VAR x: Base.Item);
VAR
  y: Base.Item;
  obj: Base.Object_;
  rel, xf, yf: INTEGER;
BEGIN

  SimpleExpression(x);

  IF (sym >= Scan.lx_eql) AND (sym <= Scan.lx_geq) THEN BEGIN

    rel := sym;
    Scan.Get(sym);
    SimpleExpression(y);
    xf := x.typ.form;
    yf := y.typ.form;

    IF x.typ = y.typ THEN BEGIN

      IF (xf IN [Base.fChar, Base.fInt]) THEN
        Gen.IntRelation(rel, x, y)
      ELSE IF xf = Base.fReal THEN
        Gen.RealRelation(rel, x, y)
      ELSE IF (xf IN [Base.fSet, Base.fPointer, Base.fProc, Base.fNilTyp, Base.fBool]) THEN BEGIN
        IF rel <= Scan.lx_neq THEN
          Gen.IntRelation(rel, x, y)
        ELSE
          Base.Mark('only = or #');
      END ELSE IF (xf = Base.fArray) AND (x.typ.base.form = Base.fChar) OR (xf = Base.fString) THEN
        Gen.StringRelation(rel, x, y)
      ELSE
        Base.Mark('illegal comparison');

    END ELSE IF (xf IN [Base.fPointer, Base.fProc]) AND (yf = Base.fNilTyp)
    OR (yf IN [Base.fPointer, Base.fProc]) AND (xf = Base.fNilTyp) THEN BEGIN

      IF rel <= Scan.lx_neq THEN
        Gen.IntRelation(rel, x,  y)
      ELSE
        Base.Mark('only = or #');

    END ELSE IF (xf = Base.fPointer) AND (yf = Base.fPointer) AND
    (IsExtension(x.typ.base, y.typ.base) OR IsExtension(y.typ.base, x.typ.base)) THEN BEGIN

      IF rel <= Scan.lx_neq THEN
        Gen.IntRelation(rel,  x, y)
      ELSE
        Base.Mark('only = or #');

    END ELSE IF (xf = Base.fArray) AND (x.typ.base.form = Base.fChar) AND
    ((yf = Base.fString) OR (yf = Base.fArray) AND (y.typ.base.form = Base.fChar))
    OR (yf = Base.fArray) AND (y.typ.base.form = Base.fChar) AND (xf = Base.fString) THEN
      Gen.StringRelation(rel, x, y)
    ELSE IF (xf = Base.fChar) AND (yf = Base.fString) AND (y.b = 2) THEN BEGIN
      Gen.StrToChar(y);
      Gen.IntRelation(rel, x, y)
    END ELSE IF (yf = Base.fChar) AND (xf = Base.fString) AND (x.b = 2) THEN BEGIN
      Gen.StrToChar(x);
      Gen.IntRelation(rel, x, y)
    END ELSE IF (xf = Base.fInt) AND (yf = Base.fInt) THEN
      Gen.IntRelation(rel,  x, y)
    ELSE
      Base.Mark('illegal comparison');
    x.typ := Base.boolType;

  END ELSE IF sym = Scan.lx_in THEN BEGIN

    Scan.Get(sym);
    CheckInt(x);
    SimpleExpression(y);
    CheckSet(y);
    Gen.In2(x, y) ;
    x.typ := Base.boolType;

  END ELSE IF sym = Scan.lx_is THEN BEGIN

    Scan.Get(sym);
    Qualident(obj);
    TypeTest(x, obj.typ, FALSE) ;
    x.typ := Base.boolType;

  END;

END;

END.

