UNIT Linker;

{$mode objfpc}{$H+}

INTERFACE

USES
  Classes, Contnrs, SysUtils, Base, Dialogs;

CONST
  versionkey = #1;
  MT         = 12;
  DescSize   = 80;

TYPE
  Module = CLASS;
  ModuleName = ARRAY [0..32-1] OF CHAR;
  TBuffer = ARRAY OF INTEGER;

  Module = CLASS
    name: ModuleName;
    next: Module;
    addr: INTEGER;
    size: INTEGER;
    body: INTEGER;
    entr: ARRAY OF INTEGER;
    desc: ARRAY OF INTEGER;
    strx: ARRAY OF INTEGER;
    vars: INTEGER;
    data: INTEGER;
    num : INTEGER;
    key : INTEGER;
    datasize: INTEGER;
  END;

VAR
  root: Module;
  AllocPtr, res, buffsize: INTEGER;
  importing, imported: ModuleName;
  modoffset: INTEGER;
  GC: TObjectList;
  MTOrg : INTEGER;

  PROCEDURE LinkBinImage(path,modname: ARRAY OF CHAR; corename: ARRAY OF CHAR);

IMPLEMENTATION

FUNCTION _DIV(CONST x, y: INTEGER): INTEGER;
VAR
  Q,R: INTEGER;
BEGIN
  Q:=TRUNC(x/y);
  R:=x MOD y;
  IF R<0 THEN BEGIN
    IF y>0 THEN
      Q:=Q-1
    ELSE
      Q:=Q+1;
  END;
  RESULT := Q;
END;

FUNCTION _MOD(CONST x, y: INTEGER): INTEGER;
BEGIN
  ASSERT(y > 0);
  IF x >= 0 THEN
    RESULT := x MOD y
  ELSE
    RESULT := (y + (x MOD y)) MOD y;
END;

PROCEDURE error(n: INTEGER; name: ARRAY OF CHAR);
BEGIN
  res := n;
  importing := name;
END;

PROCEDURE Check(s: ARRAY OF CHAR);
VAR
  i: INTEGER;
  ch: CHAR;
BEGIN
  ch := s[0];
  res := 1;
  i := 1;
  IF (ch >= 'A') AND (ch <= 'Z') OR (ch >= 'a') AND (ch <= 'z') THEN BEGIN
    REPEAT
      ch := s[i];
      INC(i);
    UNTIL NOT ((ch >= '0') AND (ch <= '9') OR (ch >= 'A') AND (ch <= 'Z')
      OR (ch >= 'a') AND (ch <= 'z') OR (ch = '.')) OR (i = 32);
    IF (i < 32) AND (ch = #0) THEN res := 0;
  END;
END;

PROCEDURE Read(VAR F: TFileStream; VAR ch: CHAR);
BEGIN
  F.Read(ch,1);
END;

PROCEDURE ReadInt(VAR F: TFileStream; VAR x: INTEGER);
BEGIN
  F.Read(x,4);
END;

PROCEDURE ReadString(VAR F: TFileStream; VAR x: ARRAY OF CHAR);
VAR
  i: INTEGER;
  ch: CHAR;
BEGIN
  i := 0;
  Read(F, ch);
  WHILE ch <> #0 DO BEGIN
    IF i < LENGTH(x)-1 THEN BEGIN
      x[i] := ch;
      INC(i)
    END;
    Read(F, ch);
  END;
  x[i] := #0;
END;

PROCEDURE ReadStringPart(VAR F: TFileStream; VAR n, cnt: INTEGER);
VAR
  ch: CHAR;
BEGIN
  n := 0;
  cnt := 0;
  Read(F, ch);
  WHILE (cnt < 4) AND (ch <> #0) DO BEGIN
    INC(n, (ORD(ch) SHL (cnt * 8)));
    INC(cnt);
    IF cnt < 4 THEN Read(F, ch);
  END;
END;

PROCEDURE LoadBin(path, name: ARRAY OF CHAR; VAR buffer: TBuffer; VAR newmod: Module);
VAR
  amod, impmod: Module;
  i, n, key, impkey, mno, nofimps, size: INTEGER;
  p, w: INTEGER;
  ch: CHAR;
  fixorgP, fixorgD, fixorgT: INTEGER;
  disp, adr, inst, pno, vno, dest, offset: INTEGER;
  name1, impname: ModuleName;
  F : TFileStream;
  import: ARRAY [0..16-1] OF Module;
BEGIN

  amod := root;
  res := 0;
  nofimps := 0;

  WHILE (amod <> NIL) AND (name <> amod.name) DO amod := amod.next;

  IF amod = NIL THEN BEGIN(*load*)

    Check(name);
    F := TFileStream.Create(path+name+'.rsc',fmOpenRead);
    IF FileExists(path+name+'.rsc') THEN BEGIN
      F.Seek(0,soFromBeginning);
      ReadString(F, name1);
      ReadInt(F, key);
      Read(F, ch);
      ReadInt(F, size);
      importing := name1;
      IF ch = versionkey THEN BEGIN
        ReadString(F, impname);   (*imports*)
        WHILE (impname[0] <> #0) AND (res = 0) DO BEGIN
          ReadInt(F, impkey);
          LoadBin(path,impname, buffer, impmod);
          import[nofimps] := impmod;
          importing := name1;
          IF res = 0 THEN BEGIN
            IF impmod.key = impkey THEN BEGIN
              INC(nofimps)
            END ELSE BEGIN
              error(3, name1);
              imported := impname;
            END;
          END;
          ReadString(F, impname)
        END;
      END ELSE
        error(2, name1);
    END ELSE
      error(1, name);

    IF res = 0 THEN BEGIN(*search for a hole in the list allocate and link*)
      amod := Module.Create;
      GC.Add(amod);
      amod.next := root;
      root := amod;
      IF amod.next = NIL THEN BEGIN
        amod.num  := 0;
        amod.addr := modoffset;
        amod.data := 0;
      END ELSE BEGIN
        amod.num  := amod.next.num + 1;
        amod.addr := amod.next.addr + amod.next.size;
        amod.data := amod.next.data + amod.next.datasize;
      END;
    END;

    IF res = 0 THEN BEGIN     (*read file*)

      amod.name     := COPY(name,1,length(name));
      amod.key      := key;
      amod.datasize := 0;

      ReadInt(F, n);          (*type descriptors*)
      i := 0;
      IF n > 0 THEN SetLength(amod.desc, n DIV 4);
      WHILE n > 0 DO BEGIN
        ReadInt(F, w);
        amod.desc[i] := w;
        INC(amod.datasize,4);
        INC(i);
        DEC(n, 4);
        INC(MTOrg,4);
      END;

      ReadInt(F, n);          (*variable space*)
      amod.vars := 0;
      IF n>0 THEN BEGIN
        amod.vars := n;
        INC(MTOrg,n);
      END;
      INC(amod.datasize,n);

      ReadInt(F, n);          (*strings*)
      n := n DIV 4;
      IF n > 0 THEN SetLength(amod.strx,n);
      i := 0;
      WHILE n > 0 DO BEGIN
        ReadInt(F, w);
        amod.strx[i] := w;
        INC(amod.datasize,4);
        INC(i);
        DEC(n);
        INC(MTOrg,4);
      END;

      ReadInt(F, n);
      amod.size := n;
      p := amod.addr;
      inc(buffsize,n);
      SetLength(buffer,buffsize+1);
      F.Read(buffer[p],n*4);  (*program code*)

      ReadString(F, name1);   (*commands*)
      WHILE name1[0] <> #0 DO BEGIN
        F.Read(w,4);
        ReadString(F, name1);
      END;

      ReadInt(F, n);          (* entries *)
      i := 0;
      if n > 0 then SetLength(amod.entr, n);
      WHILE n > 0 DO BEGIN
        ReadInt(F, w);
        amod.entr[i] := w;
        INC(i);
        DEC(n);
      END;

      n := F.position;
      ReadInt(F, w);
      i := 0;
      WHILE w >= 0 DO BEGIN
        INC(i);
        ReadInt(F, w)
      END;

      ReadInt(F, fixorgP);
      ReadInt(F, fixorgD);
      ReadInt(F, fixorgT);
      ReadInt(F, w);
      amod.body := amod.addr + _DIV(w,4);

      Read(F, ch);
      IF ch <> 'O' THEN BEGIN (*corrupted file*)
        amod := NIL;
        error(4, name)
      END;
    END;

    IF res = 0 THEN BEGIN

      adr := amod.addr + fixorgP; (*fixup of BL*)
      WHILE adr <> amod.addr DO BEGIN
        inst   := buffer[adr];
        mno    := _MOD(_DIV(inst,$100000),$10);
        pno    := _MOD(_DIV(inst,$1000),$100);
        disp   := _MOD(inst,$1000);
        impmod := import[mno-1];
        dest   := _DIV(impmod.entr[pno],4) + impmod.addr;
        offset :=  dest - adr - 1;
        buffer[adr] := _MOD(offset,$1000000) OR $0F7000000;
        adr := adr - disp;
      END;

      adr := amod.addr + fixorgD; (*fixup of LDR/STR/ADD*)
      WHILE adr <> amod.addr DO BEGIN
        inst := buffer[adr];
        mno := _MOD(_DIV(inst,$100000),$10);
        disp := _MOD(inst,$1000);
        IF mno = 0 THEN (*global*)
          buffer[adr] := (_DIV(inst,$1000000) * $10 + MT) * $100000 + amod.num * 4
        ELSE BEGIN (*import*)
          impmod := import[mno-1];
          buffer[adr] := (_DIV(inst,$1000000) * $10 + MT) * $100000 + impmod.num * 4;
          inst := buffer[adr + 1];
          vno := _MOD(inst,$100);
          offset := impmod.entr[vno];
          IF ODD(_DIV(inst,$100)) THEN
            offset := offset + impmod.addr - impmod.data;
          buffer[adr + 1] := _DIV(inst,$10000) * $10000 + offset;
        END;
        adr := adr - disp;
      END;

      adr := amod.data + fixorgT;
      WHILE adr <> amod.data DO BEGIN
        inst := amod.desc[adr-amod.data];
        mno := _MOD(_DIV(inst,$1000000),$10);
        vno := _MOD(_DIV(inst,$1000),$1000);
        disp := _MOD(inst,$1000);
        IF mno = 0 THEN (*global*)
          inst := amod.data + vno
        ELSE BEGIN (*import*)
          impmod := import[mno-1];
          offset := impmod.entr[vno];
          inst   := impmod.data + offset
        END;
        amod.desc[adr-amod.data] := inst;
        adr := adr - disp;
      END;

      (*
      adr := amod.data + fixorgT;
      i := fixorgT;
      WHILE adr <> amod.data DO BEGIN
        inst := amod.desc[i];
        mno := _MOD(_DIV(inst,$1000000),$10);
        vno := _MOD(_DIV(inst,$1000),$1000);
        disp := _MOD(inst,$1000);
        IF mno = 0 THEN (*global*)
          inst := amod.data + vno
        ELSE BEGIN (*import*)
          impmod := import[mno-1];
          offset := impmod.entr[vno];
          inst   := impmod.data + offset
        END;
        amod.desc[i] := inst;
        i   := i   - disp;
        adr := adr - disp;
      END;
      *)
    END ELSE IF res = 3 THEN BEGIN
      importing := name;
      WHILE nofimps > 0 DO DEC(nofimps);
    END;

    F.free;
  END;

  newmod := amod;

END;

PROCEDURE LinkBinImage(path,modname: ARRAY OF CHAR; corename: ARRAY OF CHAR);
CONST
  U     = $2000;
  R0    = 0;  R1    = 1;
  MT    = 12; SB    = 13; SP    = 14; LNK   = 15;
  opMov = 0;  opLsl = 1;  opAsr = 2;  opRor = 3;
  opAdd = 8;  opLdr = 8;  opSub = 9;  opStr = 10;
  opBR  = 0;  opBL  = 3;

VAR
  buff  : TBuffer;
  F     : TFileStream;
  i,w,e : INTEGER;
  amod  : Module;
  fpath : Base.TStrings;

  PROCEDURE Put0(op, a, b, c: LONGINT);
  BEGIN
    w :=  ((a*$10 + b) * $10 + op) * $10000 + c;
    F.WriteDWord(w);
    INC(buffsize);
  END;

  PROCEDURE Put1(op, a, b, im: LONGINT);
  BEGIN
    w := (((a+$40) * $10 + b) * $10 + op) * $10000 + _MOD(im,$10000);
    F.WriteDWord(w);
    INC(buffsize);
  END;

  PROCEDURE Put2(op, a, b, off: LONGINT);
  BEGIN
    w := ((op * $10 + a) * $10 + b) * $100000 + _MOD(off,$100000);
    F.WriteDWord(w);
    INC(buffsize);
  END;

  PROCEDURE Put3(op, cond, off: LONGINT);
  BEGIN
    w := ((op+12) * $10 + cond) * $1000000 + _MOD(off,$1000000);
    F.WriteDWord(w);
    INC(buffsize);
  END;

  PROCEDURE GenBL(VAR r: Module);
  BEGIN
    IF r.next <> NIL THEN GenBL(r.next);
    Put3(opBL, 7, r.body-buffsize-1);
    SetLength(r.entr, 0);
  END;

BEGIN

  GC := TObjectList.Create;
  root := nil;
  MTOrg := 0;

  modoffset := 5;
  buffsize  := modoffset;

  SetLength(buff,buffsize);
  fpath := path+Base.objdir;
  LoadBin(fpath,modname, buff, amod);

  buff[0] :=  $E7FFFFFF;// Trap or New() see Kernel.mod;

  IF res = 0 THEN BEGIN

    F := TFileStream.Create(path+corename+Base.binext,fmCreate);
    F.Seek(3*4,soFromBeginning); //for entry mtorg mtlen
    F.Write(buff[0],buffsize*4);

    IF root<>NIL THEN BEGIN

      e := buffsize;

      Put1(opSub,SP, SP,4);
      Put2(opStr,LNK,SP,0);

      amod := root;
      WHILE amod <> NIL DO BEGIN

        IF length(amod.desc)>0 THEN BEGIN
          Put1(opMov,R0,R0,amod.data);
          FOR i :=0 TO length(amod.desc)-1 DO BEGIN
            Put1(opMov,R1,R1,amod.desc[i] SHR 16);
            Put1(opLsl, R1, R1, 16);
            Put1(opAdd,R1,R1,amod.desc[i]);
            Put2(opStr,R1,R0,i*4);
          END;
        END;

        IF length(amod.strx)>0 THEN BEGIN
          Put1(opMov,R0,R0,amod.data+length(amod.desc)*4+amod.vars);
          FOR i :=0 TO length(amod.strx)-1 DO BEGIN
            Put1(opMov,R1,R1,amod.strx[i] SHR 16);
            Put1(opLsl, R1, R1, 16);
            Put1(opAdd,R1,R1,amod.strx[i]);
            Put2(opStr,R1,R0,i*4);
          END;
        END;

        amod := amod.next;
      END;

      Put1(opMov,MT,MT,MTOrg);
      amod := root;
      i := 0;
      WHILE amod <> NIL DO BEGIN
        Put1(opMov,R0,R0,amod.data);
        Put2(opStr,R0,MT,amod.num*4);
        amod := amod.next;
        inc(i);
      END;

      GenBL(root);

      Put2(opLdr,LNK,SP,0);
      Put1(opAdd,SP, SP,4);
      Put3(opBR, 7,  LNK);

      F.Seek(0,soFromBeginning);
      F.Write(e,4);
      F.Write(MTOrg,4);
      F.Write(i,4);
    END;

    F.free;

  END;
  SetLength(buff,0);
  GC.free;

END;

END.

