UNIT Parser;

{$mode objfpc}{$H+}

INTERFACE

USES
  Base, Scan, Expr, Gen, Func, Files, SysUtils, Dialogs;

VAR
  newSF: BOOLEAN;

  PROCEDURE Compile(AName: Base.TStrings; TextAddr: Pointer; Key: INTEGER);
  PROCEDURE Types(VAR typ: Base.Type_);
  PROCEDURE ProcedureType(ptype: Base.Type_; VAR parblksize: LONGINT);

IMPLEMENTATION

USES
  Main;

VAR
  modid: Base.TIdent;

PROCEDURE IdentList(clas: INTEGER; VAR first: Base.Object_);
VAR
  obj: Base.Object_;
BEGIN
  obj := NIL;
  IF sym = Scan.lx_ident THEN BEGIN
    Base.NewObj(first, Scan.id, clas);
    Scan.Get(sym);
    Expr.CheckExport(first.expo);
    WHILE sym = Scan.lx_comma DO BEGIN
      Scan.Get(sym);
      IF sym = Scan.lx_ident THEN BEGIN
        Base.NewObj(obj, Scan.id, clas);
        Scan.Get(sym);
        Expr.CheckExport(obj.expo);
      END ELSE
        Base.Mark('ident?');
    END;
    IF sym = Scan.lx_colon THEN
      Scan.Get(sym)
    ELSE
      Base.Mark(':?');
  END ELSE
    first := NIL;
END;

PROCEDURE ArrayType(VAR tp: Base.Type_);
VAR
  x: Base.Item;
  typ: Base.Type_;
  len: LONGINT;
BEGIN

  Base.NewType(typ,Base.fNoTyp,0);

  Expr.Expression(x);
  IF (x.mode = Base.cConst) AND (x.typ.form = Base.fInt) AND (x.a >= 0) THEN
    len := x.a
  ELSE BEGIN
    len := 1;
    Base.Mark('not a valid length');
  END;
  IF sym = Scan.lx_of THEN BEGIN
    Scan.Get(sym);
    Types(typ.base);
    IF (typ.base.form = Base.fArray) AND (typ.base.len < 0) THEN
      Base.Mark('dyn array not allowed');
  END ELSE IF sym = Scan.lx_comma THEN BEGIN
    Scan.Get(sym);
    ArrayType(typ.base)
  END ELSE BEGIN
    Base.Mark('missing OF');
    typ.base := Base.intType;
  END;
  typ.size := _DIV(len * typ.base.size + 3,4) * 4;
  typ.form := Base.fArray;
  typ.len  := len;
  tp       := typ;
END;

PROCEDURE RecordType(VAR t: Base.Type_);
VAR
  obj, obj0, anew, bot, bas: Base.Object_;
  typ, tp: Base.Type_;
  offset, off, n: LONGINT;
BEGIN

  tp := NIL;
  Base.NewType(typ,Base.fNoTyp,0);
  typ.mno    := -level;
  typ.nofpar := 0;

  offset     := 0;
  bot        := NIL;

  IF sym = Scan.lx_lparen THEN BEGIN
    Scan.Get(sym);
    IF Expr.level <> 0 THEN
      Base.Mark('extension of local types not implemented');
    IF sym = Scan.lx_ident THEN BEGIN
      Expr.Qualident(bas);
      IF bas.clas = Base.cTyp THEN BEGIN
        IF bas.typ.form = Base.fRecord THEN
          typ.base := bas.typ
        ELSE BEGIN
          typ.base := Base.intType;
          Base.Mark('invalid extension');
        END ;
        typ.nofpar := typ.base.nofpar + 1;
        bot := typ.base.dsc;
        offset := typ.base.size;
      END ELSE
        Base.Mark('type expected');
    END ELSE
      Base.Mark('ident expected');
    Expr.Check(Scan.lx_rparen, 'no )');
  END ;

  WHILE sym = Scan.lx_ident DO BEGIN
    n := 0;
    obj := bot;
    WHILE sym = Scan.lx_ident DO BEGIN
      obj0 := obj;
      WHILE (obj0 <> NIL) AND (obj0.name <> Scan.id) DO obj0 := obj0.next;
      IF obj0 <> NIL THEN Base.Mark('mult def');

      anew      := Base.Object_.Create;
      anew.name :=Scan.id;
      anew.clas := Base.cFld;
      anew.next := obj;
      obj       := anew;
      Base.objlist.Add(obj);

      INC(n);
      Scan.Get(sym);
      Expr.CheckExport(anew.expo);
      IF (sym <> Scan.lx_comma) AND (sym <> Scan.lx_colon) THEN
        Base.Mark('comma expected')
      ELSE IF sym = Scan.lx_comma THEN
        Scan.Get(sym);
    END;
    Check(Scan.lx_colon, 'colon expected');
    Types(tp);
    IF (tp.form = Base.fArray) AND (tp.len < 0) THEN
      Base.Mark('dyn array not allowed');
    IF tp.size > 1 THEN offset := _DIV(offset+3, 4) * 4;
    offset := offset + n * tp.size;
    off := offset;
    obj0 := obj;
    WHILE obj0 <> bot DO BEGIN
      obj0.typ := tp;
      obj0.lev := 0;
      off := off - tp.size;
      obj0.val := off;
      obj0 := obj0.next;
    END;
    bot := obj;
    IF sym = Scan.lx_semicolon THEN
      Scan.Get(sym)
    ELSE IF sym <> Scan.lx_end THEN
      Base.Mark(' ; or END');
  END;

  typ.form := Base.fRecord;
  typ.dsc  := bot;
  typ.size := _DIV(offset + 3,4) * 4;
  t        := typ;
END;

PROCEDURE FormalType(VAR typ: Base.Type_; dim: INTEGER);
VAR
  obj: Base.Object_;
  dmy: LONGINT;
BEGIN
  obj := NIL;
  IF sym = Scan.lx_ident THEN BEGIN
    Expr.Qualident(obj);
    IF obj.clas = Base.cTyp THEN
      typ := obj.typ
    ELSE BEGIN
      Base.Mark('not a type');
      typ := Base.intType;
    END;
  END ELSE IF sym = Scan.lx_array THEN BEGIN
    Scan.Get(sym);
    Expr.Check(Scan.lx_of, 'OF ?');
    IF dim >= 1 THEN Base.Mark('multi-dimensional open arrays not implemented');

    Base.NewType(typ,Base.fArray,2*Base.WordSize);
    typ.len  := -1;

    typ.len := -1;
    FormalType(typ.base,dim+1);
  END ELSE IF sym = Scan.lx_procedure THEN BEGIN
    Scan.Get(sym);
    Base.OpenScope;
    Base.NewType(typ,Base.fProc,Base.WordSize);
    dmy := 0;
    ProcedureType(typ, dmy);
    typ.dsc := Base.topScope.next;
    Base.CloseScope;
  END ELSE BEGIN
    Base.Mark('identifier expected');
    typ := Base.noType;
  END;
END;

PROCEDURE FPSection(VAR adr: LONGINT; VAR nofpar: INTEGER);
VAR
  obj, first: Base.Object_;
  tp: Base.Type_;
  parsize: LONGINT;
  cl: INTEGER;
  rdo: BOOLEAN;
BEGIN

  obj   := NIL;
  first := NIL;
  tp    := NIL;

  IF sym = Scan.lx_var THEN BEGIN
    Scan.Get(sym);
    cl := Base.cPar
  END ELSE
    cl := Base.cVar;

  IdentList(cl, first);
  FormalType(tp, 0);
  rdo := FALSE;

  IF (cl = Base.cVar) AND (tp.form >= Base.fArray) THEN BEGIN
    cl := Base.cPar;
    rdo := TRUE;
  END;

  IF (tp.form = Base.fArray) AND (tp.len < 0) OR (tp.form = Base.fRecord) THEN
    parsize := 2*Base.WordSize
  ELSE
    parsize := Base.WordSize;

  obj := first;

  WHILE obj <> NIL DO BEGIN
    INC(nofpar);
    obj.clas := cl;
    obj.typ  := tp;
    obj.rdo  := rdo;
    obj.lev  := level;
    obj.val  := adr;
    adr      := adr + parsize;
    obj      := obj.next;
  END;

  IF adr >= 52 THEN Base.Mark('too many parameters');

END;

PROCEDURE ProcedureType(ptype: Base.Type_; VAR parblksize: LONGINT);
VAR
  obj: Base.Object_;
  size: LONGINT;
  nofpar: INTEGER;
BEGIN

  obj := NIL;
  ptype.base := Base.noType;
  size := parblksize;
  nofpar := 0;
  ptype.dsc := NIL;

  IF sym = Scan.lx_lparen THEN BEGIN
    Scan.Get(sym);
    IF sym = Scan.lx_rparen THEN
      Scan.Get(sym)
    ELSE BEGIN
      FPSection(size, nofpar);
      WHILE sym = Scan.lx_semicolon DO BEGIN
        Scan.Get(sym);
        FPSection(size, nofpar)
      END;
      Expr.Check(Scan.lx_rparen, 'no )');
    END;
    IF sym = Scan.lx_colon THEN BEGIN
      Scan.Get(sym);
      IF sym = Scan.lx_ident THEN BEGIN
        Expr.Qualident(obj);
        ptype.base := obj.typ;
        IF NOT ((obj.clas = Base.cTyp) AND (obj.typ.form IN [Base.fByte .. Base.fPointer, Base.fProc])) THEN
          Base.Mark('illegal function type');
      END ELSE
        Base.Mark('type identifier expected');
    END;
  END;
  ptype.nofpar := nofpar;
  parblksize := size;

END;

PROCEDURE Types(VAR typ: Base.Type_);
VAR
  dmy: LONGINT;
  obj: Base.Object_;
  ptbase: Expr.PtrBase;
BEGIN
  obj := NIL;
  typ := Base.intType;

  IF (sym <> Scan.lx_ident) AND (sym < Scan.lx_array) THEN BEGIN
    Base.Mark('not a type');
    REPEAT
      Scan.Get(sym)
    UNTIL (sym = Scan.lx_ident) OR (sym >= Scan.lx_array)
  END;

  IF sym = Scan.lx_ident THEN BEGIN
    Expr.Qualident(obj);
    IF obj.clas = Base.cTyp THEN BEGIN
      IF (obj.typ <> NIL) AND (obj.typ.form <> Base.fNoTyp) THEN typ := obj.typ;
    END ELSE
      Base.Mark('not a type or undefined');
  END ELSE IF sym = Scan.lx_array THEN BEGIN
    Scan.Get(sym);
    ArrayType(typ);
  END ELSE IF sym = Scan.lx_record THEN BEGIN
    Scan.Get(sym);
    RecordType(typ);
    Expr.Check(Scan.lx_end, 'no END');
  END ELSE IF sym = Scan.lx_pointer THEN BEGIN
    Scan.Get(sym);
    Expr.Check(Scan.lx_to, 'no TO');

    Base.NewType(typ,Base.fPointer,Base.WordSize);
    typ.base := Base.intType;

    IF sym = Scan.lx_ident THEN BEGIN
      obj := Base.ThisObj(Scan.id);
      IF obj <> NIL THEN BEGIN
        IF (obj.clas = Base.cTyp) AND (obj.typ.form IN [Base.fRecord, Base.fNoTyp]) THEN BEGIN
          Expr.CheckRecLevel(obj.lev);
          typ.base := obj.typ;
        END ELSE IF obj.clas = Base.cMod THEN
          Base.Mark('external base type not implemented')
        ELSE
          Base.Mark('no valid base type');
      END ELSE BEGIN
        Expr.CheckRecLevel(Expr.level);
        ptbase      := Expr.PtrBase.Create;
        ptbase.name := Scan.id;
        ptbase.typ  := typ;
        ptbase.col  := Base.charerr;
        ptbase.row  := Base.lineerr;
        ptbase.next := Expr.pbsList;
        Expr.pbsList     := ptbase;
        Base.objlist.Add(ptbase);
      END ;
      Scan.Get(sym)
    END ELSE BEGIN
      Types(typ.base);
      IF typ.base.form <> Base.fRecord THEN Base.Mark('must point to record');
      Expr.CheckRecLevel(Expr.level)
    END;
  END ELSE IF sym = Scan.lx_procedure THEN BEGIN
    Scan.Get(sym);
    Base.OpenScope;
    Base.NewType(typ,Base.fProc,Base.WordSize);
    dmy := 0;
    ProcedureType(typ, dmy);
    typ.dsc := Base.topScope.next;
    Base.CloseScope;
  END ELSE
    Base.Mark('illegal type');
END;

PROCEDURE StatSequence;
VAR
  obj: Base.Object_;
  orgtype: Base.Type_;
  x, y, z, w: Base.Item;
  L0, L1, rx: LONGINT;

  PROCEDURE TypeCase(obj: Base.Object_; VAR x: Base.Item);
  VAR
    typobj: Base.Object_;
  BEGIN
    IF sym = Scan.lx_ident THEN BEGIN
      Expr.Qualident(typobj);
      Base.MakeItem(x, obj, level);
      IF typobj.clas <> Base.cTyp THEN Base.Mark('not a type');
      Expr.TypeTest(x, typobj.typ, FALSE);
      obj.typ := typobj.typ;
      Gen.CFJump(x);
      Check(Scan.lx_colon, ': expected'); StatSequence
    END ELSE BEGIN
      Gen.CFJump(x);
      Base.Mark('type id expected')
    END;
   END;

  PROCEDURE SkipCase;
  BEGIN
    WHILE sym <> Scan.lx_colon DO Scan.Get(sym);
    Scan.Get(sym);
    StatSequence;
  END;

BEGIN
  REPEAT
    obj := NIL;

    IF NOT ((sym = Scan.lx_ident) OR (sym >= Scan.lx_if) AND (sym <= Scan.lx_for) OR (sym >= Scan.lx_semicolon)) THEN BEGIN
      Base.Mark('statement expected');
      REPEAT
        Scan.Get(sym)
      UNTIL (sym = Scan.lx_ident) OR (sym >= Scan.lx_if)
    END;

    IF sym = Scan.lx_ident THEN BEGIN

      Expr.Qualident(obj);
      Base.MakeItem(x, obj, level);
      IF x.mode = Base.cSProc THEN
        Func.StandProc(obj.val)
      ELSE BEGIN
        Expr.Selector(x);
        IF sym = Scan.lx_becomes THEN BEGIN
          Scan.Get(sym);
          Expr.CheckReadOnly(x);
          Expr.Expression(y);
          IF Expr.CompTypes(x.typ, y.typ, FALSE) THEN BEGIN
            IF (x.typ.form <= Base.fPointer) OR (x.typ.form = Base.fProc) THEN
              Gen.Store(x, y)
            ELSE
              Gen.StoreStruct(x, y);
          END ELSE IF (x.typ.form = Base.fArray) AND (y.typ.form = Base.fArray) AND (x.typ.base = y.typ.base) AND (y.typ.len < 0) THEN
            Gen.StoreStruct(x, y)
          ELSE IF (x.typ.form = Base.fArray) AND (x.typ.base.form = Base.fChar) AND (y.typ.form = Base.fString) THEN
            Gen.CopyString(x, y)
          ELSE IF (x.typ.form = Base.fInt) AND (y.typ.form = Base.fInt) THEN
            Gen.Store(x, y)
          ELSE IF (x.typ.form = Base.fChar) AND (y.typ.form = Base.fString) AND (y.b = 2) THEN BEGIN
            Gen.StrToChar(y);
            Gen.Store(x, y)
          END ELSE
            Base.Mark('illegal assignment');
        END ELSE IF sym = Scan.lx_eql THEN BEGIN
          Base.Mark('should be :=');
          Scan.Get(sym);
          Expr.Expression(y)
        END ELSE IF sym = Scan.lx_lparen THEN BEGIN
          Scan.Get(sym);
          IF (x.typ.form = Base.fProc) AND (x.typ.base.form = Base.fNoTyp) THEN BEGIN
            Gen.PrepCall(x, rx);
            Expr.ParamList(x);
            Gen.Call(x, rx)
          END ELSE BEGIN
            Base.Mark('not a procedure');
            Expr.ParamList(x);
          END;
        END ELSE IF x.typ.form = Base.fProc THEN BEGIN
          IF x.typ.nofpar > 0 THEN Base.Mark('missing parameters');
          IF x.typ.base.form = Base.fNoTyp THEN BEGIN
            Gen.PrepCall(x, rx);
            Gen.Call(x, rx)
          END ELSE
            Base.Mark('not a procedure');
        END ELSE IF x.mode = Base.cTyp THEN
          Base.Mark('illegal assignment')
        ELSE
          Base.Mark('not a procedure');
      END;

    END ELSE IF sym = Scan.lx_if THEN BEGIN

      Scan.Get(sym);
      Expr.Expression(x);
      Expr.CheckBool(x);
      Gen.CFJump(x);
      Expr.Check(Scan.lx_then, 'no THEN');
      StatSequence;
      L0 := 0;
      WHILE sym = Scan.lx_elsif DO BEGIN
        Scan.Get(sym);
        Gen.FJump(L0);
        Gen.Fixup(x);
        Expr.Expression(x);
        Expr.CheckBool(x);
        Gen.CFJump(x);
        Expr.Check(Scan.lx_then, 'no THEN');
        StatSequence;
      END;
      IF sym = Scan.lx_else THEN BEGIN
        Scan.Get(sym);
        Gen.FJump(L0);
        Gen.Fixup(x);
        StatSequence;
      END ELSE
        Gen.Fixup(x);
      Gen.FixLink(L0);
      Expr.Check(Scan.lx_end, 'no END')

    END ELSE IF sym = Scan.lx_while THEN BEGIN

      Scan.Get(sym);
      L0 := Gen.Here();
      Expr.Expression(x);
      Expr.CheckBool(x);
      Gen.CFJump(x);
      Expr.Check(Scan.lx_do, 'no DO');
      StatSequence;
      Gen.BJump(L0);
      WHILE sym = Scan.lx_elsif DO BEGIN
        Scan.Get(sym);
        Gen.Fixup(x);
        Expr.Expression(x);
        Expr.CheckBool(x);
        Gen.CFJump(x);
        Expr.Check(Scan.lx_do, 'no DO');
        StatSequence;
        Gen.BJump(L0)
      END;
      Gen.Fixup(x);
      Expr.Check(Scan.lx_end, 'no END')

    END ELSE IF sym = Scan.lx_repeat THEN BEGIN

      Scan.Get(sym);
      L0 := Gen.Here();
      StatSequence;
      IF sym = Scan.lx_until THEN BEGIN
        Scan.Get(sym);
        Expr.Expression(x);
        Expr.CheckBool(x);
        Gen.CBJump(x, L0)
      END ELSE
        Base.Mark('missing UNTIL');

    END ELSE IF sym = Scan.lx_for THEN BEGIN

      Scan.Get(sym);
      IF sym = Scan.lx_ident THEN BEGIN
        Expr.Qualident(obj);
        Base.MakeItem(x, obj, level);
        Expr.CheckInt(x);
        Expr.CheckReadOnly(x);
        IF sym = Scan.lx_becomes THEN BEGIN
          Scan.Get(sym);
          Expr.Expression(y);
          Expr.CheckInt(y);
          Gen.For0(x, y);
          L0 := Gen.Here();
          Expr.Check(Scan.lx_to, 'no TO');
          Expr.Expression(z);
          Expr.CheckInt(z);
          obj.rdo := TRUE;
          IF sym = Scan.lx_by THEN BEGIN
            Scan.Get(sym);
            Expr.Expression(w);
            Expr.CheckConst(w);
            Expr.CheckInt(w)
          END ELSE
            Base.MakeConstItem(w, Base.intType, 1);
          Expr.Check(Scan.lx_do, 'no DO');
          Gen.For1(x, y, z, w, L1);
          StatSequence;
          Expr.Check(Scan.lx_end, 'no END');
          Gen.For2(x, y, w);
          Gen.BJump(L0);
          Gen.FixLink(L1);
          obj.rdo := FALSE;
        END ELSE
          Base.Mark(':= expected')
      END ELSE
        Base.Mark('identifier expected');

    END ELSE IF sym = Scan.lx_case THEN BEGIN

      Scan.Get(sym);
      IF sym = Scan.lx_ident THEN BEGIN
        Expr.Qualident(obj);
        orgtype := obj.typ;
        IF (orgtype.form = Base.fPointer) OR (orgtype.form = Base.fRecord) AND (obj.clas = Base.cPar) THEN BEGIN
          Expr.Check(Scan.lx_of, 'OF expected');
          TypeCase(obj, x);
          L0 := 0;
          WHILE sym = Scan.lx_bar DO BEGIN
            Scan.Get(sym);
            Gen.FJump(L0);
            Gen.Fixup(x);
            obj.typ := orgtype;
            TypeCase(obj, x);
          END;
          Gen.Fixup(x);
          Gen.FixLink(L0);
          obj.typ := orgtype
        END ELSE BEGIN
          Base.Mark('numeric case not implemented');
          Expr.Check(Scan.lx_of, 'OF expected');
          SkipCase;
          WHILE sym = Scan.lx_bar DO SkipCase;
        END;
      END ELSE
        Base.Mark('ident expected');
      Expr.Check(Scan.lx_end, 'no END');

    END;

    Gen.CheckRegs;
    IF sym = Scan.lx_semicolon THEN
      Scan.Get(sym)
    ELSE IF sym < Scan.lx_semicolon THEN
      Base.Mark('missing semicolon?');

  UNTIL sym > Scan.lx_semicolon;

END;

PROCEDURE Declarations(VAR varsize: LONGINT);
VAR
  obj, first: Base.Object_;
  x: Base.Item;
  tp: Base.Type_;
  ptbase: Expr.PtrBase;
  expo: BOOLEAN;
  id: Base.TIdent;
BEGIN

  Expr.pbsList := NIL;
  obj  := NIL;
  first:= NIL;
  tp   := NIL;

  IF (sym < Scan.lx_const) AND (sym <> Scan.lx_end) AND (sym <> Scan.lx_return) THEN BEGIN
    Base.Mark('declaration?');
    REPEAT
      Scan.Get(sym)
    UNTIL (sym >= Scan.lx_const) OR (sym = Scan.lx_end) OR (sym = Scan.lx_return)
  END ;

  IF sym = Scan.lx_const THEN BEGIN
    Scan.Get(sym);
    WHILE sym = Scan.lx_ident DO BEGIN
      id := Scan.id;
      Scan.Get(sym);
      Expr.CheckExport(expo);
      IF sym = Scan.lx_eql THEN
        Scan.Get(sym)
      ELSE
        Base.Mark('= ?');
      Expr.Expression(x);
      IF (x.typ.form = Base.fString) AND (x.b = 2) THEN Gen.StrToChar(x);
      Base.NewObj(obj, id, Base.cConst);
      obj.expo := expo;
      IF x.mode = Base.cConst THEN BEGIN
        obj.val := x.a;
        obj.lev := x.b;
        obj.typ := x.typ;
      END ELSE BEGIN
        Base.Mark('expression not constant');
        obj.typ := Base.intType;
      END;
      Expr.Check(Scan.lx_semicolon, '; missing');
    END;
  END;

  IF sym = Scan.lx_type THEN BEGIN
    Scan.Get(sym);
    WHILE sym = Scan.lx_ident DO BEGIN
      id := Scan.id;
      Scan.Get(sym);
      Expr.CheckExport(expo);
      IF sym = Scan.lx_eql THEN
        Scan.Get(sym)
      ELSE
        Base.Mark('=?');
      Types(tp);
      Base.NewObj(obj, id, Base.cTyp);
      obj.typ  := tp;
      obj.expo := expo;
      obj.lev  := Expr.level;
      IF tp.typobj = NIL THEN tp.typobj := obj;
      IF expo AND (obj.typ.form = Base.fRecord) THEN BEGIN
        obj.exno := exno;
        INC(exno);
      END ELSE
        obj.exno := 0;
      IF tp.form = Base.fRecord THEN BEGIN
        ptbase := Expr.pbsList;
        WHILE ptbase <> NIL DO BEGIN
          IF obj.name = ptbase.name THEN ptbase.typ.base := obj.typ;
          ptbase := ptbase.next;
        END;
        IF Expr.level = 0 THEN Gen.BuildTD(tp, dc);
      END;
      Expr.Check(Scan.lx_semicolon, '; missing');
    END;
  END;

  IF sym = Scan.lx_var THEN BEGIN
    Scan.Get(sym);
    WHILE sym = Scan.lx_ident DO BEGIN
      IdentList(Base.cVar, first);
      Types(tp);
      obj := first;
      WHILE obj <> NIL DO BEGIN
        obj.typ := tp;
        obj.lev := Expr.level;
        IF tp.size > 1 THEN varsize := _DIV(varsize + 3,4) * 4;
        obj.val := varsize;
        varsize := varsize + obj.typ.size;
        IF obj.expo THEN BEGIN
          obj.exno := exno;
          INC(exno)
        END;
        obj := obj.next;
      END ;
      Expr.Check(Scan.lx_semicolon, '; missing');
    END;
  END;

  varsize := _DIV(varsize + 3,4) * 4;
  ptbase  := Expr.pbsList;

  WHILE ptbase <> NIL DO BEGIN
    IF ptbase.typ.base.form = Base.fInt THEN BEGIN
      Base.charerr := ptbase.col;
      Base.lineerr := ptbase.row;
      Base.Mark('undefined pointer base of');
    END;
    ptbase := ptbase.next;
  END ;

  IF (sym >= Scan.lx_const) AND (sym <= Scan.lx_var) THEN
    Base.Mark('declaration in bad order');
END;

PROCEDURE ProcedureDecl;
VAR
  proc: Base.Object_;
  typ: Base.Type_;
  procid: Base.TIdent;
  x: Base.Item;
  locblksize, parblksize, L: LONGINT;
  int: BOOLEAN;
BEGIN

  proc:= NIL;
  typ := NIL;

  int := FALSE;
  Scan.Get(sym);

  IF sym = Scan.lx_times THEN BEGIN
    Scan.Get(sym);
    int := TRUE;
  END;

  IF sym = Scan.lx_ident THEN BEGIN
    procid := Scan.id;
    Scan.Get(sym);
    Base.NewObj(proc, Scan.id, Base.cConst);
    parblksize := 4;
    Base.NewType(typ,Base.fProc,Base.WordSize);
    proc.typ := typ;
    Expr.CheckExport(proc.expo);
    IF proc.expo THEN BEGIN
      proc.exno := exno;
      INC(exno);
    END;

    Base.OpenScope;
    INC(Expr.level);
    proc.val := -1;
    typ.base := Base.noType;
    ProcedureType(typ, parblksize);
    Check(Scan.lx_semicolon, 'no ;');
    locblksize := parblksize;

    Declarations(locblksize);

    proc.val := Gen.Here() * 4;
    proc.typ.dsc := Base.topScope.next;

    IF sym = Scan.lx_procedure THEN BEGIN
      L := 0;
      Gen.FJump(L);
      REPEAT
        ProcedureDecl;
        Expr.Check(Scan.lx_semicolon, 'no ;')
      UNTIL sym <> Scan.lx_procedure;
      Gen.FixLink(L);
      proc.val     := Gen.Here() * 4;
      proc.typ.dsc := Base.topScope.next;
    END ;

    Gen.Enter(parblksize, locblksize, int);

    IF sym = Scan.lx_begin THEN BEGIN
      Scan.Get(sym);
      StatSequence;
    END;

    IF sym = Scan.lx_return THEN BEGIN
      Scan.Get(sym);
      Expr.Expression(x);
      IF typ.base = Base.noType THEN
        Base.Mark('this is not a function')
      ELSE IF NOT Expr.CompTypes(typ.base, x.typ, FALSE) THEN
        Base.Mark('wrong result type');
    END ELSE IF typ.base.form <> Base.fNoTyp THEN BEGIN
      Base.Mark('function without result');
      typ.base := Base.noType;
    END;

    Gen.Return(typ.base.form, x, locblksize, int);

    Base.CloseScope;
    DEC(Expr.level);
    Check(Scan.lx_end, 'no END');

    IF sym = Scan.lx_ident THEN BEGIN
      IF Scan.id <> procid THEN Base.Mark('no match');
      Scan.Get(sym);
    END ELSE
      Base.Mark('no proc id')
  END;

  int := FALSE;

END;

PROCEDURE Module(AName: Base.TStrings; key: LONGINT);
VAR
  impid, impid1: Base.TIdent;
BEGIN

  Scan.Get(sym);
  IF sym = Scan.lx_module THEN BEGIN

    Scan.Get(sym);
    IF sym = Scan.lx_times THEN BEGIN
      Expr.version := 0;
      Scan.Get(sym);
    END ELSE
      Expr.version := 1;

    Base.OpenScope;

    IF sym = Scan.lx_ident THEN BEGIN
      modid := Scan.id;
      Scan.Get(sym);
    END ELSE
      Base.Mark('identifier expected');

    Expr.Check(Scan.lx_semicolon, 'no ;');

    Expr.level := 0;
    Expr.dc := 0;
    Expr.exno := 1;

    IF sym = Scan.lx_import THEN BEGIN

      Scan.Get(sym);

      WHILE sym = Scan.lx_ident DO BEGIN
        impid := Scan.id;
        Scan.Get(sym);
        IF sym = Scan.lx_becomes THEN BEGIN
          Scan.Get(sym);
          IF sym = Scan.lx_ident THEN BEGIN
            impid1 := Scan.id;
            Scan.Get(sym);
          END ELSE
            Base.Mark('id expected');
        END ELSE
          impid1 := impid;

        IF impid1 = 'SYSTEM' THEN
          Base.InitSystem(impid, impid1)
        ELSE
          Files.Import(AName,impid, impid1);

        IF sym = Scan.lx_comma THEN
          Scan.Get(sym)
        ELSE IF sym = Scan.lx_ident THEN
          Base.Mark('comma missing');
      END;

      Expr.Check(Scan.lx_semicolon, 'no ;');

    END;

    Gen.Open(version);
    Declarations(Expr.dc);
    Gen.SetDataSize((Expr.dc + 3) DIV 4 * 4);

    WHILE sym = Scan.lx_procedure DO BEGIN
      ProcedureDecl;
      Expr.Check(Scan.lx_semicolon, 'no ;')
    END ;
    Gen.Header;

    IF sym = Scan.lx_begin THEN BEGIN
      Scan.Get(sym);
      StatSequence;
    END ;
    Expr.Check(Scan.lx_end, 'no END');

    IF sym = Scan.lx_ident THEN BEGIN
      IF Scan.id <> modid THEN Base.Mark('no match');
      Scan.Get(sym);
    END ELSE
      Base.Mark('identifier missing');

    IF sym <> Scan.lx_period THEN Base.Mark('period missing');

    IF (Base.errcnt = 0) AND (version <> 0) THEN BEGIN
      Files.Export(AName,modid, newSF, key);
      IF newSF THEN Form1.msgview.AddMessage(modid + ' new symbol file ');
    END;

    IF Base.errcnt = 0 THEN BEGIN
      Gen.Close(AName, modid, key, exno);
      IF newSF THEN Form1.msgview.Append(IntToStr(Gen.pc)+' '+IntToStr(dc)+' '+IntToHex(key,8));
    END ELSE BEGIN
      Form1.msgview.AddMessage('compilation FAILED')
    END;

    Base.CloseScope;
    Expr.pbsList := NIL;

  END ELSE
    Base.Mark('must start with MODULE');

END;

PROCEDURE Compile(AName: Base.TStrings; TextAddr: Pointer; Key: INTEGER);
BEGIN
  Base.Init;
  Scan.Init(AName,TextAddr);
  Module(AName,Key);
  Base.Reset;
END;

END.

