UNIT Scan;

{$mode objfpc}{$H+}

INTERFACE

USES
  Base, Forms;

CONST
  lx_null = 0; lx_times = 1; lx_rdiv = 2; lx_div = 3; lx_mod = 4;
  lx_and = 5; lx_plus = 6; lx_minus = 7; lx_or = 8; lx_eql = 9;
  lx_neq = 10; lx_lss = 11; lx_leq = 12; lx_gtr = 13; lx_geq = 14;
  lx_in = 15; lx_is = 16; lx_arrow = 17; lx_period = 18;
  lx_char = 20; lx_int = 21; lx_real = 22; lx_false = 23; lx_true = 24;
  lx_nil = 25; lx_string = 26; lx_not = 27; lx_lparen = 28; lx_lbrak = 29;
  lx_lbrace = 30; lx_ident = 31;
  lx_if = 32; lx_while = 34; lx_repeat = 35; lx_case = 36; lx_for = 37;
  lx_comma = 40; lx_colon = 41; lx_becomes = 42; lx_upto = 43; lx_rparen = 44;
  lx_rbrak = 45; lx_rbrace = 46; lx_then = 47; lx_of = 48; lx_do = 49;
  lx_to = 50; lx_by = 51; lx_semicolon = 52; lx_end = 53; lx_bar = 54;
  lx_else = 55; lx_elsif = 56; lx_until = 57; lx_return = 58;
  lx_array = 60; lx_record = 61; lx_pointer = 62; lx_const = 63; lx_type = 64;
  lx_var = 65; lx_procedure = 66; lx_begin = 67; lx_import = 68; lx_module = 69;
  lx_eot = 70;

VAR
  id  : Base.TIdent;
  ival: LONGINT;
  rval: SINGLE;
  slen: LONGINT;
  str : ARRAY [0..Base.BufSize-1] OF CHAR;

  PROCEDURE Get(VAR sym: INTEGER);
  PROCEDURE Init(AName: Base.TStrings; TextAddr: Pointer);

  PROCEDURE Push(sym: INTEGER);
  PROCEDURE Pop(VAR sym: INTEGER);
  FUNCTION  Find(name: Base.TIdent): BOOLEAN;
  PROCEDURE Add(name: Base.TIdent);
  PROCEDURE Clear;

IMPLEMENTATION

TYPE
  List = CLASS
    name       : Base.TIdent;
    next       : List;
  END;

  Data = CLASS
    ch         : CHAR;
    id         : Base.TIdent;
    eot        : BOOLEAN;
    ival       : LONGINT;
    rval       : SINGLE;
    slen       : LONGINT;
    str        : ARRAY [0..Base.BufSize-1] OF CHAR;
    sym        : INTEGER;
    srcfile    : Pointer;
    srcadr     : Pointer;
    charpos    : INTEGER;
    linepos    : INTEGER;
    next       : Data;
  END;

CONST
  NKW = 34;

VAR
  ch    : CHAR;
  k     : INTEGER;
  eot   : BOOLEAN;
  KWX   : ARRAY [0..10-1] OF INTEGER;
  keyTab: ARRAY [0..NKW-1] OF RECORD
    sym: INTEGER;
    id : ARRAY [0..12-1] OF CHAR;
  END;
  srcfile    : Pointer;
  srcadr     : Pointer;
  charpos    : INTEGER;
  linepos    : INTEGER;
  scandata   : Data;
  namelist   : List;

PROCEDURE Push(sym: INTEGER);
VAR
  x: Data;
BEGIN
  x          := Data.Create;
  x.ch       := ch;
  x.id       := id;
  x.eot      := eot;
  x.sym      := sym;
  x.ival     := ival;
  x.rval     := rval;
  x.slen     := slen;
  x.str      := str;
  x.charpos  := charpos;
  x.linepos  := linepos;
  x.srcfile  := srcfile;
  x.srcadr   := srcadr;
  x.next     := scandata;
  scandata   := x;
END;

PROCEDURE Pop(VAR sym: INTEGER);
VAR
  x: Data;
BEGIN
  IF scandata <> NIL THEN BEGIN
    ch       := scandata.ch;
    id       := scandata.id;
    eot      := scandata.eot;
    sym      := scandata.sym;
    ival     := scandata.ival;
    rval     := scandata.rval;
    slen     := scandata.slen;
    str      := scandata.str;
    charpos  := scandata.charpos;
    linepos  := scandata.linepos;
    srcfile  := scandata.srcfile;
    srcadr   := scandata.srcadr;
    x        := scandata;
    scandata := x.next;
    x.free;
  END
END;

PROCEDURE Clear;

  PROCEDURE DataClear(VAR x: Data);
  BEGIN
    IF x=NIL THEN EXIT;
    IF x.next<>NIL THEN DataClear(x.next);
    x.free;
    x := NIL;
  END;

  PROCEDURE ListClear(VAR x: List);
  BEGIN
    IF x=NIL THEN EXIT;
    IF x.next<>NIL THEN ListClear(x.next);
    x.free;
    x := NIL;
  END;

BEGIN
  IF scandata <> NIL THEN DataClear(scandata);
  IF namelist <> NIL THEN ListClear(namelist);
END;

PROCEDURE Add(name: Base.TIdent);
VAR
  x: List;
BEGIN
  IF namelist <> NIL THEN BEGIN
    x := List.Create;
    x.name := name;
    x.next := namelist;
    namelist := x;
  END ELSE BEGIN
    namelist := List.Create;
    namelist.name := name;
    scandata.next := NIL;
  END;
END;

FUNCTION Find(name: Base.TIdent): BOOLEAN;
VAR
  x: List;
BEGIN
  RESULT := name = '';
  IF namelist <> NIL THEN BEGIN
    x := namelist;
    WHILE (x<>NIL) AND NOT RESULT DO BEGIN
      RESULT := x.name=name;
      x := x.next;
    END;
  END;
END;

PROCEDURE Read(VAR ch: CHAR);
BEGIN
  ch:=char(srcfile^);
  IF ch=#0 THEN BEGIN
    eot := TRUE;
  END ELSE BEGIN
    IF ch=#$09 THEN   INC(charpos,8-1);
    IF ch=#$0D THEN       charpos:=0;
    IF ch=#$0A THEN BEGIN charpos:=0; INC(linepos); END;
    INC(charpos);
    //Base.filepos := srcfile - srcadr;
    IF ch<>#$0A THEN INC(Base.filepos);
    INC(srcfile);
  END;
END;

PROCEDURE Init(AName: Base.TStrings; TextAddr: Pointer);
BEGIN
  Base.current := AName;
  srcadr       := TextAddr;
  srcfile      := TextAddr;
  charpos      := 0;
  linepos      := 0;
  Base.errpos  := 0;
  Base.errcnt  := 0;
  Base.filepos := 0;
  Base.charerr := charpos;
  Base.lineerr := linepos;
  eot := FALSE;
  Read(ch);
END;

PROCEDURE Identifier(VAR sym: INTEGER);
VAR
  i, k: INTEGER;
BEGIN
  i := 0;
  REPEAT
    IF i < Base.IdLen-1 THEN BEGIN id[i] := ch; INC(i) END ;
    Read(ch);
  UNTIL (ch < '0') OR (ch > '9') AND (ch < 'A') OR (ch > 'Z') AND (ch < 'a') OR (ch > 'z');
  id[i] := #0;
  IF i < 10 THEN BEGIN
    k := KWX[i-1];
    WHILE (id <> keyTab[k].id) AND (k < KWX[i]) DO INC(k);
    IF k < KWX[i] THEN sym := keyTab[k].sym ELSE sym := lx_ident;
  END ELSE
    sym := lx_ident;
END;

PROCEDURE Strings;
VAR
  i: INTEGER;
BEGIN
  i := 0;
  Read(ch);
  WHILE NOT eot AND (ch <> '"') DO BEGIN
    IF ch >= ' ' THEN BEGIN
      IF i < Base.BufSize-1 THEN BEGIN
        str[i] := ch;
        INC(i)
      END ELSE
        Base.Mark('string too long');
    END;
    Read(ch);
  END ;
  str[i] := #0;
  INC(i);
  Read(ch);
  slen := i;
END;

PROCEDURE HexString;
VAR
  i, m, n: INTEGER;
BEGIN
  i := 0;
  Read(ch);
  WHILE NOT eot AND (ch <> '$') DO BEGIN
    WHILE (ch = ' ') OR (ch = #9) OR (ch = #$0D) OR (ch = #$0A) DO Read(ch);
    IF      ('0' <= ch) AND (ch <= '9') THEN m := ORD(ch) - $30
    ELSE IF ('A' <= ch) AND (ch <= 'F') THEN m := ORD(ch) - $37
    ELSE BEGIN
      m := 0;
      Base.Mark('hexdig expected');
    END;
    Read(ch);
    IF      ('0' <= ch) AND (ch <= '9') THEN n := ORD(ch) - $30
    ELSE IF ('A' <= ch) AND (ch <= 'F') THEN n := ORD(ch) - $37
    ELSE BEGIN
      n := 0;
      Base.Mark('hexdig expected');
    END;
    IF i < Base.BufSize THEN BEGIN
      str[i] := CHR(m*$10 + n);
      INC(i);
    END ELSE
      Base.Mark('string too long');
    Read(ch);
  END;
  Read(ch);
  slen := i;
END;

PROCEDURE Comment;
BEGIN
  Read(ch);
  REPEAT
    WHILE NOT eot AND (ch <> '*') DO BEGIN
      IF ch = '(' THEN BEGIN
        Read(ch);
        IF ch = '*' THEN Comment;
      END ELSE
        Read(ch);
    END;
    WHILE ch = '*' DO Read(ch)
  UNTIL (ch = ')') OR eot;
  IF NOT eot THEN
    Read(ch)
  ELSE
    Base.Mark('unterminated comment');
END;

FUNCTION Ten(e: LONGINT): SINGLE;
VAR
  x, t: SINGLE;
BEGIN
  x := 1.0;
  t := 10.0;
  WHILE e > 0 DO BEGIN
    IF ODD(e) THEN x := t * x;
    t := t * t;
    e := _DIV(e,2);
  END ;
  RESULT := x;
END;

PROCEDURE Number(VAR sym: INTEGER);
CONST
  max    = 2147483647;
  maxExp = 38;
VAR
  i, k, e, n, s, h: LONGINT;
  x: SINGLE;
  d: ARRAY [0..16-1] OF INTEGER;
  negE: BOOLEAN;
BEGIN

  ival := 0;
  i := 0;
  n := 0;
  k := 0;

  REPEAT
    IF n < 16 THEN BEGIN
      d[n] := ORD(ch)-$30;
      INC(n)
    END ELSE BEGIN
      Base.Mark('too many digits');
      n := 0
    END;
    Read(ch)
  UNTIL (ch < '0') OR (ch > '9') AND (ch < 'A') OR (ch > 'F');

  IF (ch = 'H') OR (ch = 'R') OR (ch = 'X') THEN BEGIN

    REPEAT
      h := d[i];
      IF h >= 10 THEN h := h-7;
      k := k*$10 + h;
      INC(i);
    UNTIL i = n;

    IF ch = 'X' THEN BEGIN
      sym := lx_char;
      IF k < $100 THEN
        ival := k
      ELSE BEGIN
        Base.Mark('illegal value');
        ival := 0
      END;
    END ELSE IF ch = 'R' THEN BEGIN
      sym  := lx_real;
      rval := FLT(k);
    END ELSE BEGIN
      sym := lx_int;
      ival := k;
    END;

    Read(ch);

  END ELSE IF ch = '.' THEN BEGIN

    Read(ch);

    IF ch = '.' THEN BEGIN

      ch := #$7F;
      REPEAT
        IF d[i] < 10 THEN BEGIN
          IF k <= _DIV(max-d[i],10) THEN
            k := k *10 + d[i]
          ELSE BEGIN
            Base.Mark('too large');
            k := 0
          END;
        END ELSE
          Base.Mark('bad integer');
        INC(i);
      UNTIL i = n;
      sym := lx_int;
      ival := k;

    END ELSE BEGIN

      x := 0.0;
      e := 0;

      REPEAT
        x := x * 10.0 + FLT(d[i]);
        INC(i)
      UNTIL i = n;

      WHILE (ch >= '0') AND (ch <= '9') DO BEGIN
        x := x * 10.0 + FLT(ORD(ch) - $30);
        DEC(e);
        Read(ch);
      END;

      IF (ch = 'E') OR (ch = 'D') THEN BEGIN

        Read(ch);

        s := 0;
        IF ch = '-' THEN BEGIN
          negE := TRUE;
          Read(ch);
        END ELSE BEGIN
          negE := FALSE;
          IF ch = '+' THEN Read(ch);
        END;

        IF (ch >= '0') AND (ch <= '9') THEN BEGIN
          REPEAT
            s := s*10 + ORD(ch)-$30;
            Read(ch);
          UNTIL (ch < '0') OR (ch >'9');
          IF negE THEN e := e-s ELSE e := e+s;
        END ELSE
          Base.Mark('digit?');

      END ;

      IF e < 0 THEN BEGIN
        IF e >= -maxExp THEN
          x := x / Ten(-e)
        ELSE
          x := 0.0;
      END ELSE IF e > 0 THEN BEGIN
        IF e <= maxExp THEN
          x := Ten(e) * x
        ELSE BEGIN
          x := 0.0;
          Base.Mark('too large')
        END;
      END;
      sym := lx_real;
      rval := x;

    END;

  END ELSE BEGIN

    REPEAT
      IF d[i] < 10 THEN BEGIN
        IF k <= _DIV(max-d[i],10) THEN
          k := k*10 + d[i]
        ELSE BEGIN
          Base.Mark('too large');
          k := 0
        END;
      END ELSE
        Base.Mark('bad integer');
      INC(i);
    UNTIL i = n;

    sym := lx_int;
    ival := k;

  END;
END;

PROCEDURE Get(VAR sym: INTEGER);
BEGIN
  REPEAT

    Application.ProcessMessages;

    WHILE NOT eot AND (ch <= ' ') DO Read(ch);

    Base.charerr := charpos;
    Base.lineerr := linepos;

    IF eot THEN
      sym := lx_eot
    ELSE IF ch < 'A' THEN BEGIN
      IF ch < '0' THEN BEGIN
        IF          ch = '"' THEN BEGIN Strings;   sym := lx_string
        END ELSE IF ch = '#' THEN BEGIN Read(ch);  sym := lx_neq
        END ELSE IF ch = '$' THEN BEGIN HexString; sym := lx_string
        END ELSE IF ch = '&' THEN BEGIN Read(ch);  sym := lx_and
        END ELSE IF ch = ')' THEN BEGIN Read(ch);  sym := lx_rparen
        END ELSE IF ch = '*' THEN BEGIN Read(ch);  sym := lx_times
        END ELSE IF ch = '+' THEN BEGIN Read(ch);  sym := lx_plus
        END ELSE IF ch = ',' THEN BEGIN Read(ch);  sym := lx_comma
        END ELSE IF ch = '-' THEN BEGIN Read(ch);  sym := lx_minus
        END ELSE IF ch = '/' THEN BEGIN Read(ch);  sym := lx_rdiv
        END ELSE IF ch = '(' THEN BEGIN
          Read(ch);
          IF ch = '*' THEN BEGIN
            sym := lx_null;
            Comment
          END ELSE
            sym := lx_lparen;
        END ELSE IF ch = '.' THEN BEGIN
          Read(ch);
          IF ch = '.' THEN BEGIN
            Read(ch);
            sym := lx_upto;
          END ELSE
            sym := lx_period;
        END ELSE BEGIN
          Read(ch);
          sym := lx_null;
        END;
      END ELSE IF ch < ':' THEN
        Number(sym)
      ELSE IF ch = ':' THEN BEGIN
        Read(ch);
        IF ch = '=' THEN BEGIN
          Read(ch);
          sym := lx_becomes;
        END ELSE
          sym := lx_colon;
      END ELSE IF ch = ';' THEN BEGIN
        Read(ch);
        sym := lx_semicolon
      END ELSE IF ch = '<' THEN BEGIN
        Read(ch);
        IF ch = '=' THEN BEGIN
          Read(ch);
          sym := lx_leq;
        END ELSE
          sym := lx_lss;
      END ELSE IF ch = '=' THEN BEGIN
        Read(ch);
        sym := lx_eql;
      END ELSE IF ch = '>' THEN BEGIN
        Read(ch);
        IF ch = '=' THEN BEGIN
          Read(ch);
          sym := lx_geq;
        END ELSE
          sym := lx_gtr;
      END ELSE BEGIN
        Read(ch);
        sym := lx_null;
      END;
    END ELSE IF ch < '[' THEN
      Identifier(sym)
    ELSE IF ch < 'a' THEN BEGIN
      IF      ch = '[' THEN sym := lx_lbrak
      ELSE IF ch = ']' THEN sym := lx_rbrak
      ELSE IF ch = '^' THEN sym := lx_arrow
      ELSE                  sym := lx_null;
      Read(ch);
    END ELSE IF ch < '{' THEN
      Identifier(sym)
    ELSE BEGIN
      IF      ch = '{'  THEN sym := lx_lbrace
      ELSE IF ch = '}'  THEN sym := lx_rbrace
      ELSE IF ch = '|'  THEN sym := lx_bar
      ELSE IF ch = '~'  THEN sym := lx_not
      ELSE IF ch = #$7F THEN sym := lx_upto
      ELSE                   sym := lx_null;
      Read(ch);
    END;
  UNTIL sym <> lx_null;
END;

PROCEDURE EnterKW(sym: INTEGER; name: ARRAY OF CHAR);
BEGIN
  keyTab[k].id := COPY(name,0,length(name));
  keyTab[k].sym := sym;
  INC(k)
END;

BEGIN
  k := 0;
  KWX[0] := 0;
  KWX[1] := 0;
  EnterKW(lx_if, 'IF');
  EnterKW(lx_do, 'DO');
  EnterKW(lx_of, 'OF');
  EnterKW(lx_or, 'OR');
  EnterKW(lx_to, 'TO');
  EnterKW(lx_in, 'IN');
  EnterKW(lx_is, 'IS');
  EnterKW(lx_by, 'BY');
  KWX[2] := k;
  EnterKW(lx_end, 'END');
  EnterKW(lx_nil, 'NIL');
  EnterKW(lx_var, 'VAR');
  EnterKW(lx_div, 'DIV');
  EnterKW(lx_mod, 'MOD');
  EnterKW(lx_for, 'FOR');
  KWX[3] := k;
  EnterKW(lx_else, 'ELSE');
  EnterKW(lx_then, 'THEN');
  EnterKW(lx_true, 'TRUE');
  EnterKW(lx_type, 'TYPE');
  EnterKW(lx_case, 'CASE');
  KWX[4] := k;
  EnterKW(lx_elsif, 'ELSIF');
  EnterKW(lx_false, 'FALSE');
  EnterKW(lx_array, 'ARRAY');
  EnterKW(lx_begin, 'BEGIN');
  EnterKW(lx_const, 'CONST');
  EnterKW(lx_until, 'UNTIL');
  EnterKW(lx_while, 'WHILE');
  KWX[5] := k;
  EnterKW(lx_record, 'RECORD');
  EnterKW(lx_repeat, 'REPEAT');
  EnterKW(lx_return, 'RETURN');
  EnterKW(lx_import, 'IMPORT');
  EnterKW(lx_module, 'MODULE');
  KWX[6] := k;
  EnterKW(lx_pointer, 'POINTER');
  KWX[7] := k;
  KWX[8] := k;
  EnterKW(lx_procedure, 'PROCEDURE');
  KWX[9] := k;

  scandata := NIL;
  namelist := NIL;

END.

