unit XsDebuggers;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Controls, ExtCtrls, Forms,
  StdCtrls, Messages, LMessages, XsControls,
  Types, LCLType, LCLintf, Graphics, Dialogs;

CONST
  MemSize            = $100000;
  MemWords           = (MemSize div 4);

type
  TInstr = RECORD
    mode : byte;
    flag : boolean;
    code : STRING[8];
    mnem : STRING[4];
    r1,r2: STRING[4];
    val  : STRING[10];
    ival : longint;
  END;

  Flags = RECORD
    Z,N,C,V: BOOLEAN;
  END;

  { TXSDebugView }

  TXSDebugView = class(TXSScrollbox)
  private
    PC          : longword;
    M           : array [0..MemWords-1] of LONGWORD;
    R           : array [0..16-1] of longword;
    H           : longword;
    Flag        : Flags;
    fcode       : array of longword;
    finstr      : array of TInstr;
    maxvis      : integer;
    fstart      : integer;
    floaded     : boolean;
    frepaint    : boolean;
    factiveline : integer;
    flineheight : integer;
    fcharwidth  : integer;
    fgutwidth   : integer;
    fentry      : integer;
    fmtorg      : integer;
    fmtlen      : integer;
    fsize       : integer;
    fchanged    : integer;
    frightline  : integer;
    procedure   ScrollChange(Sender: TObject);
    PROCEDURE   Step;
    PROCEDURE   Reset;
  protected
    procedure   Paint; override;
    procedure   Resize; override;
    procedure   WMMouseWheel(var Message: TLMMouseEvent); message LM_MOUSEWHEEL;
  public
    constructor Create(TheOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   LoadFromFile(aname: string; var res: boolean);
  published

  end;

  { TXSModTable }
  TXSObjView = class;

  TXSModTable = class(TXSScrollbox)
  private
    maxvis      : integer;
    fstart      : integer;
    flinemax    : integer;
    factiveline : integer;
    flineheight : integer;
    fcharwidth  : integer;
    fwordwidth  : integer;
    fsize       : integer;
    fgutwidth   : integer;
    fparent     : TXSObjView;
    fcount      : integer;
    procedure   ScrollChange(Sender: TObject);
  protected
    procedure   Paint; override;
    procedure   Resize; override;
    procedure   WMMouseWheel(var Message: TLMMouseEvent); message LM_MOUSEWHEEL;
  public
    constructor Create(TheOwner: TComponent); override;
    destructor  Destroy; override;
  published

  end;

  { TXSObjView }

  TXSObjView = class(TCustomControl)
  private
    fbgcolor    : integer;
    ffrcolor    : integer;
    fdebugtop   : integer;
    fonmousedown: TMouseEvent;
    freglabel   : array [0..15] of TLabel;
    fregval     : array [0..15] of TLabel;
    fflaglabel  : array [0..7] of TLabel;
    fflagval    : array [0..7] of TLabel;
    fbits       : array [0..31] of TLabel;
    fnumbits    : array [0..31] of TLabel;
    fdebug      : TXSDebugView;
    fbitpanel   : TGroupBox;
    freggroup   : TGroupBox;
    fflaggroup  : TGroupBox;
    fflagwidth  : integer;
    fregwidth   : integer;
    fbitswidth  : integer;
    fstepflag   : boolean;
    fmodpanel   : TGroupBox;
    fmodtab     : TXSModTable;
    function    GetLoaded: boolean;
  protected
    procedure   Paint; override;
    procedure   Resize; override;
    procedure   MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure   KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure   DebugMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
  public
    constructor Create(TheOwner: TComponent); override;
    destructor  Destroy; override;
    function    LoadFromFile(const aname: string): boolean;
    procedure   Step;
    procedure   Reset;
  published
    property OnMouseDown: TMouseEvent read fonmousedown write fonmousedown;
    property Loaded: boolean read GetLoaded;
  end;

implementation

const
  opMOV = 0;  opLSL = 1;  opASR = 2;  opROR = 3;
  opAND = 4;  opANN = 5;  opIOR = 6;  opXOR = 7;
  opADD = 8;  opSUB = 9;  opMUL = 10; opDIV = 11;
  opFAD = 12; opFSB = 13; opFML = 14; opFDV = 15;

VAR
  mnemo0, mnemo1: ARRAY [0..16-1] OF ARRAY [0..5-1] OF CHAR;

FUNCTION _DIV(CONST x, y: INTEGER): INTEGER;
VAR
  Q,R: INTEGER;
BEGIN
  Q:=TRUNC(x/y);
  R:=x MOD y;
  IF R<0 THEN BEGIN
    IF y>0 THEN
      Q:=Q-1
    ELSE
      Q:=Q+1;
  END;
  RESULT := Q;
END;

FUNCTION _MOD(CONST x, y: INTEGER): INTEGER;
BEGIN
  ASSERT(y > 0);
  IF x >= 0 THEN
    RESULT := x MOD y
  ELSE
    RESULT := (y + (x MOD y)) MOD y;
END;

PROCEDURE WriteReg(r: LONGINT; VAR reg: SHORTSTRING);
BEGIN
  IF          r < 12 THEN BEGIN reg := 'R'+ IntToStr(_MOD(r,$10));
  END ELSE IF r = 12 THEN       reg := 'MT'
  ELSE     IF r = 13 THEN       reg := 'SB'
  ELSE     IF r = 14 THEN       reg := 'SP'
  ELSE                          reg := 'LNK';
END;

FUNCTION Opcode(w: LONGINT): TInstr;
VAR
  k, op, u, a, b, c, x: LONGINT;
  instr: TInstr;
BEGIN

  k  := _MOD(_DIV(w,$40000000),4  );
  a  := _MOD(_DIV(w,$1000000) ,$10);
  b  := _MOD(_DIV(w,$100000)  ,$10);
  op := _MOD(_DIV(w,$10000)   ,$10);
  c  := _MOD(     w,$10           );
  u  := _MOD(_DIV(w,$20000000),2  );

  instr.code := IntToHex(w,8);

  IF k = 0 THEN BEGIN
    instr.mode := 0;
    instr.mnem := mnemo0[op];
    instr.flag := u = 0;
    IF u = 1 THEN instr.mnem :=  instr.mnem + '''';
    WriteReg(a,instr.r1); WriteReg(b,instr.r2); WriteReg(c,instr.val);
    instr.ival := 0;
  END ELSE IF k = 1 THEN BEGIN
    instr.mode := 1;
    instr.mnem := mnemo0[op];
    instr.flag := u = 0;
    IF u = 1 THEN instr.mnem := instr.mnem + '''';
    WriteReg(a,instr.r1); WriteReg(b,instr.r2);
    w := _MOD(w,$10000);
    IF w >= $8000 THEN w := w - $10000;
    instr.val := IntToStr(w);
    instr.ival := w;
  END ELSE IF k = 2 THEN BEGIN (*LDR/STR*)
    instr.mode := 2;
    instr.flag := u = 0;
    IF u = 1 THEN instr.mnem := 'STR'
    ELSE          instr.mnem := 'LDR';
    WriteReg(a,instr.r1); WriteReg(b,instr.r2);
    w := _MOD(w,$100000);
    IF w >= $80000 THEN w := w - $100000;
    instr.val  := IntToStr(w);
    instr.ival := w;
  END ELSE IF k = 3 THEN BEGIN (*Branch instr*)
    instr.mode := 3;
    instr.mnem := 'B';
    IF ODD(_DIV(w ,$10000000)) THEN instr.mnem := instr.mnem + 'L';
    instr.mnem := instr.mnem + mnemo1[a];
    instr.flag := u = 0;
    IF u = 0 THEN
      WriteReg(c,instr.r1)
    ELSE BEGIN
      w := _MOD(w,$100000);
      IF w >= $80000 THEN w := w - $100000;
      instr.val  := IntToStr(w);
      instr.ival := w;
    END;
  END;

  RESULT := instr;

END;

FUNCTION b2i(b: BOOLEAN): BYTE;
BEGIN
  IF b THEN RESULT := 1 ELSE RESULT := 0;
END;

FUNCTION i2b(i: LONGWORD): BOOLEAN;
BEGIN
  IF i <> 0 THEN i2b := TRUE ELSE i2b := FALSE;
END;

function Fpadd(x, y : LONGWORD; u, v : Boolean) : LONGWORD;
var
  ys, xs : Boolean;
  xm, xe, ye, ym, e0, sum, s, e1, t3, shift : LONGWORD;
  y0, x0, x3, y3 : int32;

begin
  xs := (x and $80000000) <>  0;
  if not(u) then begin
    xe := (x shr 23) and $FF;
    xm := ((x and $7FFFFF) shl 1) or $1000000;
    if xs then x0 := -xm else x0 := xm;
  end else begin
    xe := 150;
    x0 := (((x and $00FFFFFF) shl 8) shr 7);
  end;

  ys := (y and $80000000) <> 0;
  ye := (y shr 23) and $FF;
  ym := ((y and $7FFFFF) shl 1);

  if (not(u) and not(v)) then ym := ym or $1000000;
  if ys then y0 := -ym else y0 := ym;

  if (ye > xe) then begin
    shift := ye - xe;
    e0 := ye;
    if (shift > 31) then x3 :=  (x0 shr 31) else x3 := (x0 shr shift);
    y3 := y0;
  end else begin
    shift := xe - ye;
    e0    := xe;
    x3    := x0;
    if (shift > 31) then
      y3 :=  (y0 shr 31)
    else
      y3 := (y0 shr shift);
  end;

   sum := ((b2i(xs) shl 26) or (b2i(xs) shl 25) or (x3 and $01FFFFFF))
    + ((b2i(ys) shl 26) or (b2i(ys) shl 25) or (y3 and $01FFFFFF));

  if ((sum and (1 shl 26)) <> 0) then s := -sum else s:= sum;
  s := (s + 1) and $07FFFFFF;

  e1 := e0 + 1;
  t3 := s shr 1;
  if ((s and $3FFFFFC) <> 0) then begin
    while ((t3 and (1 shl 24)) = 0) do begin
      t3 := t3 shl 1;
      dec(e1);
    end;
  end else begin
    t3 := t3 shl 24;
    e1 := e1 - 24;
  end;

  if v then begin
    Result := (sum shl 5) shr 6;
    exit;
  end else begin
    if ((x and $7FFFFFFF) = 0) then begin
      if not(u) then Result := y else Result := 0;
      exit;
    end;

    if ((y and $7FFFFFFF) = 0) then begin
      Result := x;
      exit;
    end;

    if ((t3 and $01FFFFFF) = 0) or ((e1 and $100) <> 0) then begin
      Result := 0;
      exit;
    end;

  end;

  Result :=  ((sum and $04000000) shl 5) or (e1 shl 23) or ((t3 shr 1) and $7FFFFF);

end;

function Fpmul( x,y :  LONGWORD) : LONGWORD;
var
  sign, xe, ye, xm, ym, e1, z0 : LONGWORD;
  m : QWORD;
begin
  sign := (x xor y) and $80000000;
  xe   := (x shr 23) and $FF;
  ye   := (y shr 23) and $FF;

  xm    := (x and $7FFFFF) or $800000;
  ym    := (y and $7FFFFF) or $800000;
  m     := xm * ym;

  e1 := (xe + ye) - 127;

  if ((m and (1 shl 47)) <> 0) then begin
    inc(e1);
    z0 := (m shr 24) and $7FFFFF;
  end else
    z0 := (m shr 23) and $7FFFFF;

  if ((xe = 0) or (ye = 0)) then begin
    Result := 0;
    exit;
  end;

  if ((e1 and $100) = 0) then begin
    Result :=  sign or ((e1 and $FF) shl 23) or z0;
    exit;
  end;

  if ((e1 and $80) = 0) then begin
    Result := sign or ($FF shl 23) or z0;
    exit;
  end;

  Result := 0;

end;

function Fpdiv( x,y :  LONGWORD ) : LONGWORD;
var
  e1, sign, xe, ye,
  xm, ym, q1,  q2 : LONGWORD;
begin
  sign := (x xor y)  and $80000000;
  xe   := (x shr 23) and $FF;
  ye   := (y shr 23) and $FF;
  xm   := (x and $7FFFFF) or $800000;
  ym   := (y and $7FFFFF) or $800000;
  q1   := (xm * (1 shl 23) DIV ym);
  e1   := (xe - ye) + 126;

  if ((q1 and $800000) <> 0) then begin
    inc(e1);
    q2 := q1 and $7FFFFF;
  end else
    q2 := (q1 shl 1) and $7FFFFF;

  if (xe = 0) then begin
    Result := 0;
    exit;
  end;

  if (ye = 0) then begin
    Result :=  sign or ($FF shl 23);
    exit;
  end;

  if ((e1 and $100) = 0) then begin
    Result := sign or ((e1 and $FF) shl 23) or q2;
    exit;
  end;

  if ((e1 and $80) = 0) then begin
    Result :=  sign or ($FF shl 23) or q2;
    exit;
  end;

  Result := 0;

end;

{ TXSModTable }

constructor TXSModTable.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  fparent := TXSObjView(TheOwner);
  VScroll.OnChange := @ScrollChange;
end;

destructor TXSModTable.Destroy;
begin

  inherited Destroy;
end;

procedure TXSModTable.ScrollChange(Sender: TObject);
begin
  fstart := VScroll.position;
  Invalidate;
end;

procedure TXSModTable.WMMouseWheel(var Message: TLMMouseEvent);
begin
  if Message.WheelDelta > 0 then begin
    VScroll.position := VScroll.position - 4
  end else begin
    if VScroll.position + 4 > fcount - maxvis then
      VScroll.position := fcount - maxvis
    else
      VScroll.position := VScroll.position + 4;
  end;
  fstart := VScroll.position;
  if fstart < 0 then fstart := 0;
end;

procedure TXSModTable.Resize;
begin
  inherited Resize;
  flineheight := canvas.TextHeight('A')+2;
  maxvis := height div flineheight;

  if fstart > fcount - maxvis then fstart := fcount - maxvis;
  if fstart < 0 then fstart := 0;

  VScroll.position := fstart;
  fcharwidth := canvas.TextWidth('W');
  fgutwidth := 5+fcharwidth*8+5;
  fwordwidth:= fcharwidth*8+fcharwidth;
end;

procedure TXSModTable.Paint;
var
  i,x,y,len,ftw: integer;
  txt: string;
  oldpen: TPen;
  oldbrush: TBrush;
  ww,n: integer;
begin
  inherited Paint;

  flinemax := fstart + height div flineheight - 1;

  oldpen   := Canvas.Pen;
  oldbrush := Canvas.Brush;
  ww := width-fgutwidth-fwordwidth-VScroll.Width;
  ftw := (ww div fwordwidth) + 1;

  len := 0;
  if fparent<>nil then len := fparent.fdebug.fmtorg div 4 + fparent.fdebug.fmtlen;

  if ftw <= 0 then
    fcount := len
  else
    fcount := (len div ftw) + 1;

  VScroll.Max      := fcount - 1;
  VScroll.PageSize := maxvis - 1;

  Canvas.Brush.Style := bsSolid;
  Canvas.Pen.Color   := $E5E5E5;
  Canvas.Brush.Color := $E5E5E5;
  Canvas.Rectangle(0,0,fgutwidth,height);
  Canvas.Pen.Color   := clSilver;
  Canvas.Line(fgutwidth,0,fgutwidth,height);

  Canvas.Pen.Color   := clWhite;
  Canvas.Brush.Color := clWhite;
  Canvas.FillRect(fgutwidth+1,0,width-VScroll.Width,height);

  Canvas.Brush.Style := bsClear;

  n := ftw * fstart;

  for i:=fstart to flinemax do begin

    Canvas.Font.Style := [];
    Canvas.Font.Color := clGray;
    Canvas.TextOut(5,(i-fstart)*flineheight+3,IntToHex(n*4,8));

    Canvas.Font.Style := [];
    Canvas.Font.Color := clDefault;
    if (fparent<>nil) and (ftw >= 0) then begin
      x := 0;
      while (x<=ww) and (n<len) and (len<=length(fparent.fdebug.M)) do begin
        txt := IntToHex(fparent.fdebug.M[n],8);
        if fparent.fdebug.fchanged = n then begin
          Canvas.Font.Style := [fsBold];
          Canvas.Font.Color := clRed;
        end else begin
          Canvas.Font.Style := [];
          Canvas.Font.Color := clDefault;
        end;
        if n*4 >= fparent.fdebug.fmtorg then Canvas.Font.Color := clMaroon;
        Canvas.TextOut(fgutwidth+5+x,(i-fstart)*flineheight+3,txt);
        inc(x,fwordwidth);
        inc(n);
      end;
    end;

    Canvas.Font.Style := [];
    Canvas.Font.Color := clDefault;
  end;

  Canvas.Brush := oldbrush;
  Canvas.Pen   := oldpen;

end;

{ TXSDebugView }

constructor TXSDebugView.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  fstart := 0;
  factiveline := 1;
  Font.name := 'Courier New';
  Font.Size := 9;
  HScroll.Visible := false;
  VScroll.OnChange := @ScrollChange;
  floaded := false;
  frepaint := false;
  fmtorg   := 0;
  fmtlen   := 0;
  fentry   := 0;
  fchanged := -1;
  frightline := 50;
end;

destructor TXSDebugView.Destroy;
begin
  SetLength(fcode,0);
  SetLength(finstr,0);
  inherited Destroy;
end;

procedure TXSDebugView.ScrollChange(Sender: TObject);
begin
  fstart := VScroll.position;
  Invalidate;
end;

procedure TXSDebugView.WMMouseWheel(var Message: TLMMouseEvent);
begin
  if Message.WheelDelta > 0 then begin
    VScroll.position := VScroll.position - 4
  end else begin
    if VScroll.position + 4 > length(finstr) - maxvis then
      VScroll.position := length(finstr) - maxvis
    else
      VScroll.position := VScroll.position + 4;
  end;
  fstart := VScroll.position;
  if fstart < 0 then fstart := 0;
end;

procedure TXSDebugView.LoadFromFile(aname: string; var res: boolean);
var
  F: TFileStream;
  i: integer;
begin
  res := false;

  if FileExists(aname) then begin
    res := true;
    F:= TFileStream.Create(aname,fmOpenRead);
    F.Seek(0,soBeginning);

    F.Read(fentry,4);
    F.Read(fmtorg,4);
    F.Read(fmtlen,4);

    fsize  := (F.size - F.position) DIV 4;
    SetLength(fcode,fsize);
    SetLength(finstr,fsize);
    F.Read(fcode[0],F.size-F.position);

    i := 0;
    while i < fsize do begin
      finstr[i] := Opcode(fcode[i]);
      INC(i);
    end;

    floaded := true;
    frepaint := false;

    Reset;

    F.free;
  end;
end;

PROCEDURE TXSDebugView.Reset;
VAR
  i: INTEGER;
BEGIN

  FillDWord(M[0],MemWords,0);

  PC := 0;
  IF (fentry>=0) AND (fentry<=fsize) THEN PC := fentry;

  factiveline := PC;
  FOR i := 0 TO 15 DO R[i]:=0;

  R[14] := memsize-4;
  R[15] := PC*4;
  fstart := PC;
  if frepaint then
    Invalidate
  else begin
    Repaint;
    frepaint := true;
  end;

  fchanged := -1;
END;

PROCEDURE TXSDebugView.Step;
CONST
  IO = $0FFFC0;
VAR
  k, op, u, v, a, b, c, w, im : LONGINT;
  aval,bval,cval,adr,off      : LONGWORD;
  x                           : BOOLEAN;
  tmp                         : QWORD;

  PROCEDURE SetReg(reg: INTEGER; val: LONGWORD);
  BEGIN
    R[reg]:= val;
    Flag.Z := (val = 0);
    Flag.N := LONGINT(val) < 0;
  END;

BEGIN

  w := fcode[PC];
  INC(PC);

  k  := _MOD(_DIV(w,$40000000),4  );
  u  := _MOD(_DIV(w,$20000000),2  );
  v  := _MOD(_DIV(w,$10000000),2  );
  a  := _MOD(_DIV(w,$1000000) ,$10);
  b  := _MOD(_DIV(w,$100000)  ,$10);
  op := _MOD(_DIV(w,$10000)   ,$10);
  c  := _MOD(     w,$10           );
  im := _MOD(w,     $10000        );

  aval := 0;
  bval := 0;
  cval := 0;
  adr  := 0;

  IF k IN [0,1] THEN BEGIN

    bval := R[b];

    IF k = 0 THEN
      cval:= R[c]
    ELSE IF v = 0 THEN
      cval := im
    ELSE
      cval := ($FFFF0000 OR im);

    CASE op OF

      opLSL : aval := bval SHL (cval AND 31);
      opASR : aval := (LONGWORD(bval)) SHR (cval AND 31);
      opROR : aval := (bval SHR (cval AND 31)) OR (bval SHL (-cval AND 31));
      opAND : aval := bval AND cval;
      opANN : aval := bval AND NOT(cval);
      opIOR : aval := bval OR  cval;
      opXOR : aval := bval XOR cval;

      opMOV :
              BEGIN
                IF u = 0 THEN
                  aval := cval
                ELSE IF k <>0 THEN
                  aval := (cval SHL 16)
                ELSE IF v <> 0 THEN
                  aval := $D0 OR (b2i(Flag.N) * $80000000) OR (b2i(Flag.Z) * $40000000) OR (b2i(Flag.C) * $20000000) OR (b2i(Flag.V) * $10000000)
                ELSE
                  aval := H;
              END;
      opADD :
              BEGIN
                aval := bval + cval;
                if ((u <> 0) AND Flag.C) THEN aval := aval + 1;
                Flag.C:= aval < bval;
                Flag.V:= ((NOT(bval XOR cval) AND (aval XOR bval)) SHR 31) <> 0;
              END;
      opSUB :
              BEGIN
                x := cval > bval;
                aval := bval - cval;
                if ((u <> 0) AND Flag.C) THEN aval := aval - 1;
                Flag.C := x;
                Flag.V := (((bval XOR cval) AND (aval XOR bval)) SHR 31) <> 0;
              END;
      opMUL :
              BEGIN
                IF u = 0  THEN
                  tmp := QWORD(INTEGER(bval)) * QWORD(INTEGER(cval))
                ELSE
                  tmp := QWORD(bval) * QWORD(cval);
                aval := tmp;
                H := tmp SHR 32;
              END;
      opDIV :
              BEGIN
                IF cval <= 0 THEN BEGIN
                  aval  := $DEADBEEF;
                  H     := $DEADBEEF;
                END ELSE BEGIN
                  aval := LONGINT(bval) DIV LONGINT(cval);
                  H := LONGINT(bval) MOD LONGINT(cval);
                  IF LONGINT(H) < 0 THEN BEGIN
                    DEC(aval);
                    H := H + (cval);
                  END;
                END;
              END;

      opFAD : aval := Fpadd(bval,cval,i2b(u),i2b(v));
      opFSB : aval := Fpadd(bval,(cval xor $80000000),i2b(u),i2b(v));
      opFML : aval := Fpmul(bval,cval);
      opFDV : aval := Fpdiv(bval,cval);

    END;

    SetReg(a,aval);

  END ELSE IF k = 2 THEN BEGIN (*LDR/STR*)

    adr := _DIV(R[b] + _MOD(w,$100000),4);
    IF u = 0 THEN BEGIN
      aval := M[adr];
      SetReg(a,aval);
    END ELSE BEGIN
      M[adr] := R[a];
      IF adr <= _DIV(fmtorg,4) + fmtlen THEN fchanged := adr;
    END;

  END ELSE IF k = 3 THEN BEGIN (*Branch instr*)

    CASE ((w SHR 24) AND 15) OF
      0:  x := Flag.N;
      1:  x := Flag.Z;
      2:  x := Flag.C;
      3:  x := Flag.V;
      4:  x := Flag.C OR Flag.Z;
      5:  x := Flag.N <> Flag.V;
      6:  x := (Flag.N <> Flag.V) OR Flag.Z;
      7:  x := TRUE;
      8:  x := NOT(Flag.N);
      9:  x := NOT(Flag.Z);
      10: x := NOT(Flag.C);
      11: x := NOT(Flag.V);
      12: x := NOT(Flag.C OR Flag.Z);
      13: x := NOT(Flag.N <> Flag.V);
      14: x := NOT((Flag.N <> Flag.V) OR Flag.Z);
      15: x := FALSE;
    END;

    IF x THEN BEGIN
      IF v<>0 THEN SetReg(15,PC*4);
      IF u = 0 THEN
        PC := _MOD(_DIV(R[c],4),length(M))
      ELSE BEGIN
        w := _MOD(w,$100000);
        IF w >= $80000 THEN w := w - $100000;
        PC := _MOD(PC + w,length(M));
      END;
    END;

  END;

  factiveline := PC;

  if (factiveline > fstart + maxvis -1) or (factiveline < fstart) then fstart := factiveline - 2;

END;

procedure TXSDebugView.Resize;
begin
  inherited Resize;
  flineheight := canvas.TextHeight('A')+2;
  if fstart > length(finstr) - maxvis then fstart := length(finstr) - maxvis;
  if fstart < 0 then fstart := 0;
  maxvis := height div flineheight;
  VScroll.position := fstart;
  fcharwidth := canvas.TextWidth('W');
  fgutwidth := 5+fcharwidth*8+5;
end;

procedure TXSDebugView.Paint;
var
  i,x,y,max: integer;
  txt: string;
  oldpen: TPen;
  oldbrush: TBrush;
begin
  inherited Paint;

  max := fstart + height div flineheight - 1;
  if max > length(finstr)-1 then max := length(finstr)-1;
  maxvis := height div flineheight;
  VScroll.Max      := length(finstr)-1;
  VScroll.PageSize := maxvis - 1;

  if (length(finstr)-1 > maxvis) and (fstart + maxvis > length(finstr)) then fstart := length(finstr) - maxvis;
  if fstart < 0 then fstart := 0;

  oldpen   := Canvas.Pen;
  oldbrush := Canvas.Brush;

  Canvas.Brush.Style := bsSolid;
  Canvas.Pen.Color   := $E5E5E5;
  Canvas.Brush.Color := $E5E5E5;
  Canvas.Rectangle(0,0,fgutwidth,height-1);
  Canvas.Pen.Color   := clSilver;
  Canvas.Line(fgutwidth,0,fgutwidth,height-1);

  Canvas.Brush.Style := bsClear;
  for i:=fstart to max do begin

    Canvas.Font.Style := [];
    Canvas.Font.Color := clGray;
    txt := '00000000'+IntToStr(i);
    txt := copy(txt,length(IntToStr(i))+1,8);
    Canvas.TextOut(5,(i-fstart)*flineheight+1,txt);

    if i=factiveline then begin
      Canvas.Pen.Color  := $CC8800;
      Canvas.Brush.Color:= $CC8800;
      Canvas.Font.Color := clWhite;
      Canvas.Rectangle(fgutwidth+2,(i-fstart)*flineheight,frightline*fcharwidth-1(*width-VScroll.width-1*),(i-fstart)*flineheight+flineheight);
      Canvas.Brush := oldbrush;
      Canvas.Pen   := oldpen;
    end;

    Canvas.Brush.Style := bsClear;
    Canvas.Font.Style := [];
    if i<>factiveline then Canvas.Font.Color := clDefault;
    x := fgutwidth+15;
    y := (i-fstart)*flineheight+1;
    Canvas.TextOut(x,y,finstr[i].code);

    Canvas.Font.Style := [fsBold];
    if i<>factiveline then Canvas.Font.Color := clNavy;
    inc(x,8*fcharwidth+30);
    Canvas.TextOut(x,y,finstr[i].mnem);

    Canvas.Font.Style := [];
    if i<>factiveline then Canvas.Font.Color := clDefault;

    inc(x,7*fcharwidth);
    case finstr[i].mode of
      0,1,2:
        begin
          if i<>factiveline then Canvas.Font.Color := clMaroon;
          Canvas.TextOut(x,y,finstr[i].r1);
          inc(x,4*fcharwidth);
          Canvas.TextOut(x,y,finstr[i].r2);
          if i<>factiveline then Canvas.Font.Color := clDefault;
          inc(x,4*fcharwidth);
          if finstr[i].ival < 0 then dec(x,fcharwidth);
          if finstr[i].mode = 0 then begin
            if i<>factiveline then Canvas.Font.Color := clMaroon
          end else if finstr[i].mode in [1,2] then begin
            if i<>factiveline then Canvas.Font.Color := clHighlight;
          end;
          Canvas.TextOut(x,y,finstr[i].val);
          if i<>factiveline then Canvas.Font.Color := clDefault;
        end;
      3:
        begin
          if finstr[i].flag then begin
            if i<>factiveline then Canvas.Font.Color := clMaroon;
            Canvas.TextOut(x,y,finstr[i].r1)
          end else begin
            if finstr[i].ival < 0 then dec(x,fcharwidth);
            if i<>factiveline then Canvas.Font.Color := clHighlight;
            Canvas.TextOut(x,y,finstr[i].val)
          end;
        end;
    end;

    Canvas.Font.Style := [];
    Canvas.Font.Color := clDefault;
  end;

  if i>0 then inc(i);
  Canvas.Font.Style := [];
  Canvas.Font.Color := clGray;
  while i < maxvis do begin
    txt := '00000000'+IntToStr(i);
    txt := copy(txt,length(IntToStr(i))+1,8);
    Canvas.TextOut(5,(i-fstart)*flineheight+1,txt);
    inc(i);
  end;

  Canvas.Pen.Color   := clSilver;
  Canvas.Line(frightline*fcharwidth,0,frightline*fcharwidth,height-1);
  Canvas.Line(frightline*fcharwidth+2,0,frightline*fcharwidth+2,height-1);

  Canvas.Brush := oldbrush;
  Canvas.Pen   := oldpen;

end;

{ TXSObjView }

constructor TXSObjView.Create(TheOwner: TComponent);
const
  flags = 'ZNCVpquv';
var
  i,x,y: integer;
  ch: string[1];
  rname: string[30];
begin
  inherited Create(TheOwner);

  fstepflag   := true;

  fbgcolor    := cl3DLight;
  ffrcolor    := clSilver;
  Font.Name   := 'Courier New';
  Font.Size   := 9;

  fregwidth := 4*140;
  freggroup := TGroupBox.Create(self);
  freggroup.parent := self;
  freggroup.top := 10;
  freggroup.left := 10;
  freggroup.width := fregwidth;
  freggroup.height := 4*20+20;
  freggroup.Caption:=' Registers ';
  freggroup.color := fbgcolor;

  x := 5;
  y := 0;
  ch:= ' ';
  for i:=0 to 15 do begin

    case i of
      0..11: rname := 'R'+IntToStr(i)+ch+' =';
      12: rname := 'MT  =';
      13: rname := 'SB  =';
      14: rname := 'SP  =';
      15: rname := 'LNK =';
    end;

    freglabel[i] := TLabel.Create(self);
    freglabel[i].parent := freggroup;
    freglabel[i].font.style := [fsBold];
    freglabel[i].AutoSize:=true;
    freglabel[i].Transparent:=true;
    freglabel[i].caption := rname;
    freglabel[i].left := x;
    freglabel[i].top := y*20;

    fregval[i] := TLabel.Create(self);
    fregval[i].parent := freggroup;
    fregval[i].AutoSize:=true;
    fregval[i].Transparent:=true;
    fregval[i].caption := '00000000H';
    fregval[i].left := freglabel[i].left+50;
    fregval[i].top := y*20;

    inc(y);
    if ((i+1) mod 4 = 0) then begin inc(x,140); y := 0; end;
    if i >= 9 then ch:='';
  end;

  fflagwidth := 2*45;
  fflaggroup := TGroupBox.Create(self);
  fflaggroup.parent := self;
  fflaggroup.top := 10;
  fflaggroup.left := freggroup.left+freggroup.width+10;
  fflaggroup.width := fflagwidth;
  fflaggroup.height := 4*20+20;
  fflaggroup.Caption:=' Flags ';
  fflaggroup.color := fbgcolor;

  x := 5;
  y := 0;
  ch:= ' ';
  for i:=0 to 7 do begin
    fflaglabel[i] := TLabel.Create(self);
    fflaglabel[i].parent := fflaggroup;
    fflaglabel[i].font.style := [fsBold];
    fflaglabel[i].AutoSize:=true;
    fflaglabel[i].Transparent:=true;
    fflaglabel[i].caption := flags[i+1]+' =';
    fflaglabel[i].left := x;
    fflaglabel[i].top := y*20;

    fflagval[i] := TLabel.Create(self);
    fflagval[i].parent := fflaggroup;
    fflagval[i].AutoSize:=true;
    fflagval[i].Transparent:=true;
    fflagval[i].caption := '0';
    fflagval[i].left := fflaglabel[i].left+25;
    fflagval[i].top := y*20;

    inc(y);
    if ((i+1) mod 4 = 0) then begin inc(x,45); y := 0; end;
  end;

  fbitswidth := 10+32*17+5;
  fbitpanel:= TGroupBox.Create(self);
  fbitpanel.parent := self;
  fbitpanel.left := 10;
  fbitpanel.top := fflaggroup.top+fflaggroup.height+5;
  fbitpanel.width := fbitswidth;
  fbitpanel.height := 55;
  fbitpanel.caption := ' Instruction bits ';
  fbitpanel.color := fbgcolor;

  for i:=31 downto 0 do begin
    fnumbits[i] := TLabel.Create(self);
    fnumbits[i].parent := fbitpanel;
    fnumbits[i].AutoSize:=true;
    fnumbits[i].Transparent:=true;
    if i<=9 then ch := '0' else ch := '';
    fnumbits[i].caption := ch+IntToStr(i);
    fnumbits[i].left := 5+(31-i)*17;
    fnumbits[i].top  := 0;
    fnumbits[i].enabled := false;
  end;
  for i:=0 to 31 do begin
    fbits[i] := TLabel.Create(self);
    fbits[i].parent := fbitpanel;
    fbits[i].font.style := [fsBold];
    fbits[i].AutoSize:=true;
    fbits[i].Transparent:=true;
    fbits[i].caption := '0';
    fbits[i].left := 8+i*17;
    fbits[i].top  := 15;
  end;

  fmodpanel := TGroupBox.Create(self);
  fmodpanel.parent := self;
  fmodpanel.left := 10;
  fmodpanel.top := fbitpanel.top+fbitpanel.height+5;
  fmodpanel.height := 115;
  fmodpanel.Caption := 'Modules memory table';

  fmodtab := TXSModTable.Create(self);
  fmodtab.parent := fmodpanel;
  fmodtab.align := alClient;
  fmodtab.HScroll.visible := false;
  fmodtab.BorderSpacing.Left := 3;
  fmodtab.BorderSpacing.Right := 5;
  fmodtab.BorderSpacing.Bottom := 5;

  fdebug := TXSDebugView.Create(self);
  fdebug.parent := self;
  fdebug.OnMouseDown:=@DebugMouseDown;
  fdebugtop := fmodpanel.top+fmodpanel.height+5;

end;

destructor TXSObjView.Destroy;
var
  i: integer;
begin
  for i:=0 to 31 do begin
    if i<=7 then begin
      fflaglabel[i].free;
      fflagval[i].free;
    end;
    if i<=15 then begin
      freglabel[i].free;
      fregval[i].free;
    end;
    fnumbits[i].free;
    fbits[i].free;
  end;
  fdebug.free;
  fflaggroup.free;
  freggroup.free;
  fbitpanel.free;
  fmodtab.free;
  fmodpanel.free;
  inherited Destroy;
end;

procedure TXSObjView.MouseDown(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if Assigned(fonmousedown) then fonmousedown(self,Button,Shift,X,Y);
end;

procedure TXSObjView.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
  if fdebug.floaded and (Key = VK_F8) then begin
    Step;
    Application.ProcessMessages;
  end;
end;

procedure TXSObjView.DebugMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
begin
  MouseDown(Button,Shift,X,Y);
end;

function TXSObjView.LoadFromFile(const aname: string): boolean;
begin
  fdebug.LoadFromFile(aname,result);
  Reset;
end;

procedure TXSObjView.Reset;
var
  i,y: integer;
begin
  fdebug.Reset;
  for i := 0 to 15 do fregval[i].caption := IntToHex(fdebug.R[i],8)+'H';
  y := fdebug.fcode[fdebug.PC];
  for i:=31 downto 0 do begin
    fbits[i].caption := '0';
    if ((y and 1) > 0) then begin
      fbits[i].caption := '1';
      case i of
        0: fflagval[4].caption := '1';
        1: fflagval[5].caption := '1';
        2: fflagval[6].caption := '1';
        3: fflagval[7].caption := '1';
      end;
    end else begin
      case i of
        0: fflagval[4].caption := '0';
        1: fflagval[5].caption := '0';
        2: fflagval[6].caption := '0';
        3: fflagval[7].caption := '0';
      end;
    end;
    y := y shr 1;
  end;
end;

procedure TXSObjView.Step;
var
  i,y: integer;
begin

    y := fdebug.fcode[fdebug.PC];
    for i:=31 downto 0 do begin
      fbits[i].caption := '0';
      if ((y and 1) > 0) then begin
        fbits[i].caption := '1';
        case i of
          0: fflagval[4].caption := '1';
          1: fflagval[5].caption := '1';
          2: fflagval[6].caption := '1';
          3: fflagval[7].caption := '1';
        end;
      end else begin
        case i of
          0: fflagval[4].caption := '0';
          1: fflagval[5].caption := '0';
          2: fflagval[6].caption := '0';
          3: fflagval[7].caption := '0';
        end;
      end;
      y := y shr 1;
    end;

    if fstepflag then fdebug.Step else ;//Reset;

    fflagval[0].caption := IntToStr(ORD(fdebug.Flag.Z));
    fflagval[1].caption := IntToStr(ORD(fdebug.Flag.N));
    fflagval[2].caption := IntToStr(ORD(fdebug.Flag.C));
    fflagval[3].caption := IntToStr(ORD(fdebug.Flag.V));

    for i:=0 to 15 do fregval[i].caption := IntToHex(fdebug.R[i],8)+'H';

    fstepflag := fdebug.PC < length(fdebug.fcode)-1;

    Invalidate;

end;

procedure TXSObjView.Resize;
begin
  inherited Resize;
  fdebug.BoundsRect := Rect(5+10,fdebugtop+5,width-10-5,height-10-5);
  if (width>0) and (width < freggroup.left+fregwidth+10) then freggroup.width := width-freggroup.left-10;
  if (freggroup.width < freggroup.left+fregwidth) and (width > freggroup.left+fregwidth+10) then freggroup.width := fregwidth;

  if (width>0) and (width < fflaggroup.left+fflagwidth+10) then fflaggroup.width := width-fflaggroup.left-10;
  if (fflaggroup.width < fflaggroup.left+fflagwidth) and (width > fflaggroup.left+fflagwidth+10) then fflaggroup.width := fflagwidth;

  if (width>0) and (width < fbitpanel.left+fbitswidth+10) then fbitpanel.width := width-fbitpanel.left-10;
  if (fbitpanel.width < fbitpanel.left+fbitswidth) and (width > fbitpanel.left+fbitswidth+10) then fbitpanel.width := fbitswidth;

  fmodpanel.width := width - 20;
  Invalidate;
end;

function TXSObjView.GetLoaded: boolean;
begin
  result := fdebug.floaded;
end;

procedure TXSObjView.Paint;
var
  frect: TRect;
begin
  frect := Rect(5,5,width-5,height-5);
  Canvas.Pen.Color   := ffrcolor;
  Canvas.Brush.Color := fbgcolor;
  Canvas.Roundrect(frect,5,5);

  frect := fdebug.BoundsRect;
  dec(frect.left,5);
  dec(frect.top,5);
  inc(frect.right,5);
  inc(frect.bottom,5);
  Canvas.Roundrect(frect,5,5);
end;

begin
  mnemo0[0]  := 'MOV';
  mnemo0[1]  := 'LSL';
  mnemo0[2]  := 'ASR';
  mnemo0[3]  := 'ROR';
  mnemo0[4]  := 'AND';
  mnemo0[5]  := 'ANN';
  mnemo0[6]  := 'IOR';
  mnemo0[7]  := 'XOR';
  mnemo0[8]  := 'ADD';
  mnemo0[9]  := 'SUB';
  mnemo0[10] := 'MUL';
  mnemo0[11] := 'DIV';
  mnemo0[12] := 'FAD';
  mnemo0[13] := 'FSB';
  mnemo0[14] := 'FML';
  mnemo0[15] := 'FDV';

  mnemo1[0]  := 'MI ';
  mnemo1[1]  := 'EQ ';
  mnemo1[2]  := 'LS ';
  mnemo1[5]  := 'LT ';
  mnemo1[6]  := 'LE ';
  mnemo1[8]  := 'PL ';
  mnemo1[9]  := 'NE ';
  mnemo1[10] := 'HI ';
  mnemo1[13] := 'GE ';
  mnemo1[14] := 'GT ';
  mnemo1[15] := 'NO ';
end.

